import contextlib
import os
import shutil
import ntpath
from urllib.parse import urljoin, urlparse

from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls
from dli.models import log_public_functions_calls_using, AttributesDict


def flatten_s3_file_path(s3_path):
    path = urlparse(s3_path)[2]  # [scheme, netloc, path, ...]
    head, tail = ntpath.split(path)
    file = tail or ntpath.basename(head)
    return file


def get_or_create_os_path(s3_path: str, to: str, flatten: bool) -> str:
    if flatten:
        destination_key = flatten_s3_file_path(s3_path)
    else:
        destination_key = urlparse(s3_path).path.lstrip('/')

    to_path = os.path.join(
        to, os.path.normpath(destination_key)
    )

    to_path = os.path.abspath(to_path)

    if len(to_path) > 259 and os.name == 'nt':
        raise Exception(f"Apologies {s3_path} can't be downloaded "
                        f"as the file name would be too long. You "
                        f"may want to try calling again with "
                        f"Instance.download(flatten=True), which "
                        f"will put the file in a directory of your choice")

    else:
        directory, _ = os.path.split(to_path)
        os.makedirs(directory, exist_ok=True)

    return to_path


@log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['datafile_id', 'path']
)
class FileModel(AttributesDict):

    @contextlib.contextmanager
    def open(self):
        response = self._client.session.get(
            urljoin(
                self._client._environment.consumption,
                consumption_urls.consumption_download.format(
                    id=self.datafile_id,
                    path=self.path
                )
            ),
            stream=True
        )
        # otherwise you get raw secure
        response.raw.decode_content = True
        yield response.raw
        response.close()

    def download(self, to='./', flatten=False):
        """
        Download one or more files and save them in the user specified
        directory.

        :param str to: The path on the system, where the files
            should be saved. Must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :return: The path to the directory where the files were written.
        """

        # if flatten:
        #     c_path, filename = os.path.split(to_path)
        #     pd_path, as_of_date = os.path.split(c_path)
        #     pd_path = pd_path[2:].replace('/', '-')
        #     to_path = './{}-{}/{}'.format(pd_path, as_of_date[11:], filename)
        to_path = get_or_create_os_path(
            s3_path=self.path, to=to, flatten=flatten
        )

        print(f'Downloading {self.path} to: {to_path}...')

        with self.open() as download_stream:
            with open(to_path, 'wb') as target_download:
                # copyfileobj is just a simple buffered
                # file copy function with some sane
                # defaults and optimisations.
                shutil.copyfileobj(
                    download_stream, target_download
                )
                print(f'Completed download to: {to_path}.')

        return to_path



import copy
import pytest

from dli.models.paginator import Paginator


@pytest.fixture
def empty_package_request():
    return {
            "class": ["Access Requests", "collection"],
            "properties": {"page": 1, "pages_count": 1},
            "entities": []
           }


class TestPaginator:
    def test_retrieve_zero_packages(self, test_client, empty_package_request,
                                    dataset_request):
        test_client._session.get.return_value.json.side_effect = [
            empty_package_request, dataset_request
        ]

        paginator = Paginator(
            '/',
            test_client._Package,
        )

        assert len(list(paginator)) == 0

    def test_threaded_paginator(self, test_client,
                                package_request, dataset_request):

        p = {
            'class': ['Access Requests', 'collection'],
            'properties': {'page': 1, 'pages_count': 2},
            'entities': [{
                'properties': {
                    'packageId': '8137dec0-8569-11e9-8727-f64112f300c3',
                    'name': 'Test Package',
                    'data': {
                        'name': 'Test Package',
                        'topic': 'Test Topic',
                        'keywords': ['test'],
                        'hasAccess': True,
                        'description': 'DESC',
                        'documentation': 'href',
                        'dataSensitivity': 'Private',
                    },
                    'total_records': 1,
                }
            }]
           }
        p2 = {
            'class': ['Access Requests', 'collection'],
            'properties': {'page': 2, 'pages_count': 2},
            'entities': [{
                'properties': {
                    'packageId': '8137dec0-8569-11e9-8727-f64112f300c3',
                    'name': 'Test Package 2',
                    'data': {
                        'name': 'Test Package 2',
                        'topic': 'Test Topic',
                        'keywords': ['test'],
                        'hasAccess': True,
                        'description': 'DESC',
                        'documentation': 'href',
                        'dataSensitivity': 'Private',
                    },
                    'total_records': 1,
                }
            }]
           }

        # Note: there is no side-effect call to dataset because the call that was
        # previously made in the constructor by .shape has now been made lazy to
        # avoid an explosion of calls.
        test_client._session.get.return_value.json.side_effect = [p, p2]

        paginator = Paginator(
            '/',
            test_client._Package,
            lambda a: {}
        )

        assert len(list(paginator)) == 2




import warnings
from dli.client.components import SirenComponent
from dli.client.utils import ensure_count_is_valid
from dli.client.components.urls import search_urls


class Search(SirenComponent):
    """
    Catalogue search functions.
    """

    def search(self, term, count=100):
        """
        Search across all catalogue entities given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned.

        :returns: A list of Catalogue entities
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_catalogue_entities = client.search(
                    term="Automotive",
                    count=100
                )

        """
        warnings.warn(
            f'Calls to search will be deprecated in future. '
            f'New endpoints will be provided in the Packages '
            f'and Datasets that will allow you to list ALL '
            f'that are visible to your user.',
            PendingDeprecationWarning
        )
        ensure_count_is_valid(count)
        return self.session.get(
            search_urls.search_root, params={'query': term, 'page_size': count}
        ).to_many_siren('')

    def search_packages(self, term, count=100):
        """
        Search across packages in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of packages
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_packages = client.search_packages(
                    term="Automotive",
                    count=100
                )

        """
        warnings.warn(
            f'Calls to search packages will be deprecated in future. '
            f'New endpoints will be provided in the Packages '
            f'and Datasets that will allow you to list ALL '
            f'that are visible to your user.',
            PendingDeprecationWarning
        )
        ensure_count_is_valid(count)
        return self.session.get(
            search_urls.search_packages, params={'query': term, 'page_size': count}
        ).to_many_siren('package')

    def search_datasets(self, term, count=100):
        """
        Search across datasets in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of datasets
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                results = client.search_datasets(
                    term="CDS",
                    count=100
                )

        """
        warnings.warn(
            f'Calls to search datasets will be deprecated in future. '
            f'New endpoints will be provided in the Packages '
            f'and Datasets that will allow you to list ALL '
            f'that are visible to your user.',
            PendingDeprecationWarning
        )
        ensure_count_is_valid(count)
        return self.session.get(
            search_urls.search_datasets, params={'query': term, 'page_size': count}
        ).to_many_siren('dataset')



import json
from urllib.parse import urlsplit, parse_qs
import pytest
import furl
import httpretty

from urllib.parse import urlencode
from dli.client.listener import _Listener
from tests.test_session import TestEnv


@pytest.fixture
def mock_listener_client():
    _Listener.run(debug=True, port=9999)

    with _Listener.app.test_client() as client:
        yield client

@httpretty.activate
def auth(mock_listener_client, state, sam_status='200', cat_status='200',
         headers=True, cookies=True):
    postbox = '1'

    sam_response = json.dumps({'id_token': 'ID', 'access_token': 'ACCESS'})
    httpretty.register_uri(
        httpretty.POST,
        'http://sam.local/sso/oauth2/realms/root/realms/Customers/access_token',
        status=sam_status,
        body=sam_response
    )

    cat_response = json.dumps({'access_token': 'JWT'})
    httpretty.register_uri(
        httpretty.POST,
        'http://catalogue.local/api/identity/v2/auth/token',
        status=cat_status,
        body=cat_response,
    )

    target_builder = furl.furl("/")

    if headers:
        target_builder.args = {
            'code': 'CODE',
            'state' : state,
            'client_id': 'CLIENT_ID'
        }

    session = mock_listener_client.session_transaction()
    with session as s:
        if cookies:
            s['postbox'] = '1'

        t = TestEnv()
        s["sam_client"] = t.sam_client
        s["catalogue"] = t.catalogue
        s["code_verifier"] = "test"
        s["sam"] = t.sam

    b = mock_listener_client.get(target_builder.url)
    return b


def test_listener_is_up(mock_listener_client):
    a = mock_listener_client.application.url_map
    rules = ['/shutdown', '/login']
    assert(all(k in [x.rule for x in a.iter_rules()] for k in rules))


@httpretty.activate
def test_login(mock_listener_client):
    response_text = 'Login'
    httpretty.register_uri(
        httpretty.HEAD, 'http://catalogue.local/login',
        status=302,
        body=response_text,
        adding_headers={"Location": "https://sam.local"}
    )
    postbox = 1
    a = mock_listener_client.get(
        f"/login?postbox={postbox}&{urlencode(TestEnv().__dict__)}"
    )
    parts = urlsplit(a.headers["Location"])
    assert(parts.netloc == "sam.local")

    params = parse_qs(parts.query)
    keys = ['state', 'client_id', 'response_type','redirect_uri',
     'scope','code_challenge','code_challenge_method']
    assert(all(k in params.keys() for k in keys))
    assert(params["state"][0] == str(postbox))
    assert(params["redirect_uri"][0].startswith("http://localhost"))


def test_miss_postbox_login(mock_listener_client):
    b = mock_listener_client.get("/login")
    assert(b.status_code == 403)


def test_miss_env_login(mock_listener_client):
    b = mock_listener_client.get("/login?postbox=1")
    assert(b.status_code == 500)

@httpretty.activate
def test_miss_cookie_authenticate(mock_listener_client):
    # todo this module's fixture use requires precedence
    # cookie is a side effect on mock_listener_client
    b = auth(mock_listener_client, 1, cookies=False)
    print(b.data)
    assert(b.status_code == 500)
    assert("internal error" in str(b.data.decode("utf-8")))

@httpretty.activate
def test_authenticate(mock_listener_client):
    b = auth(mock_listener_client, 1)
    assert(b.status_code == 200)
    assert("now logged in" in str(b.data.decode("utf-8")))

@httpretty.activate
def test_authenticate_get(mock_listener_client):
    b = auth(mock_listener_client, 1)
    assert(b.status_code == 200)
    assert("now logged in" in str(b.data.decode("utf-8")))
    assert(json.loads(mock_listener_client.get("/get/1").data)["jwt"] == 'JWT')

@httpretty.activate
def test_miss_headers_authenticate(mock_listener_client):
    b = auth(mock_listener_client, 1, headers=False)
    assert(b.status_code == 403)

@httpretty.activate
def test_mitm_authenticate(mock_listener_client):
    b = auth(mock_listener_client, 2)
    assert(b.status_code==403)

@httpretty.activate
def test_bad_SAM_response_authenticate(mock_listener_client):
    b = auth(mock_listener_client, 1, sam_status='403')
    assert(b.status_code==403)

@httpretty.activate
def test_bad_Catalogue_response_authenticate(mock_listener_client):
    b = auth(mock_listener_client, 1, cat_status='403')
    assert (b.status_code == 403)




import math
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Lock


class Paginator:

    def __init__(self, url, model,
         instantiation_method=None, instantiation_override=None,
         page=1, total_field=None, page_size=25, max_workers=10,
    ):
        self.page = page
        self.page_size = page_size
        self.url = url
        self.model = model
        self.instantiation_factory = instantiation_method
        self.override = instantiation_override
        self.cache = []
        self.total_field = total_field
        self.lock = Lock()
        self.max_workers = max_workers

    def _request(self, page=None):

        params={
            'page': self.page if not page else page,
        }

        if self.page_size:
            params['page_size'] = self.page_size

        return self.model._client.session.get(
            self.url,
            params=params
        ).json()

    def __iter__(self):
        if not self.cache:
            try:
                tp = ThreadPoolExecutor(max_workers=self.max_workers)

                inst = self._request()
                key = None
                if inst.get("entities", None) is not None:
                    # v1 route
                    key = "entities"
                    block = "properties"
                    count = "pages_count"
                else:
                    # v2 route
                    key = "data"
                    block = "meta"
                    count = "total_pages"

                if key:

                    def _get_and_instantiate(page):
                        inst = self._request(page)
                        self._iter_instance(inst, key)

                    self._iter_instance(inst, key)

                    if not self.total_field:
                        total_pages = inst[block][count] + 1
                    else:
                        # total field gives us how many of X
                        # so we must divide by the page_size
                        # todo this is code-debt caused by dictionary
                        # todo pagination from catalogue DL-2809
                        total_count = inst[block][self.total_field]
                        total_pages = math.ceil(total_count / self.page_size) + 1

                    rng = range(2, total_pages)

                    list(tp.map(_get_and_instantiate, rng))

            finally:
                tp.shutdown(wait=True)

        yield from self.cache

    def _iter_instance(self, inst, key):
        obj = inst if not key else inst.get(key, [])

        # we use the __init__ method of the class unless
        # (or) provided instantiation method
        # ... useful to transform data first
        # (or) provided override to control cache insertion
        # ... useful to flatten multi-item

        if not self.override:
            for f in obj:
                if not self.instantiation_factory:
                    m = self.model(f)
                else:
                    m = self.instantiation_factory(f)

                with self.lock:
                    self.cache.append(m)
        else:
            self.override(obj, self.cache)



import inspect
import functools
import io
import base64
import ipywidgets as widgets

from collections import defaultdict

from contextlib import redirect_stdout
from dli.client import session
from ipywidgets import Layout
import IPython  # noqa: I900

from dli.models.dataset_model import DatasetModel
from dli.models.package_model import PackageModel


def help(obj):
    print("This is a " + type(obj).__name__ + "\n" + "*"*20,
        "\nIt has these functions:\n" + "\n".join(
          ["." + method_name for method_name in dir(obj)
                  if callable(getattr(obj, method_name))
           and ((not method_name.startswith("_")) or method_name=='__iter__')])
        ,
        "\n\nIt has these properties:\n" + "\n".join(
          ["." + method_name for method_name in dir(obj)
                  if not callable(getattr(obj, method_name))
           and ((not method_name.startswith("_")) or method_name=='__iter__')])
    + "\n\nAnd it looks like this:\n" + str(obj)
         )


def redirect(fun):
    with io.StringIO() as buf, redirect_stdout(buf):
        fun()
        output = buf.getvalue()
        return output


class SelectorWidget:
    def __init__(self, ipython):
        self.ipython = ipython
        self._name_counts = defaultdict(lambda: 0)
        self.packages = []
        self.datasets = []
        self.current_obj = None
        self.dl = None

        self.which_packages = widgets.RadioButtons(
            options=['Only Mine', 'All Packages'],
            disabled=False
        )

        self.button = widgets.Button(
            description='Connect',
            disabled=False,
            button_style='',
            tooltip='Click me',
            icon='check'
        )

        self.button.on_click(self.btn_eventhandler)

        self.copy_button = widgets.Button(
            description='Copy current object',
            disabled=False,
            button_style='',
            tooltip='Click me',
            icon='check',
            layout=Layout(width='100%', height='30px')
        )
        self.copy_button.on_click(self.btn_copyhandler)

        self.api_key = widgets.Text(
            value='',
            placeholder='Enter your api key',
            description='Enter your api key',
            disabled=False,
            layout=Layout(width='100%', height='30px')
        )

        self.env_dd = widgets.Dropdown(
            options=[('Prod', 1), ('UAT', 2), ('QA', 3)],
            value=1,
            description='Environment',
            disabled=False,
            layout=Layout(width='100%')
        )

        self.package_dd = widgets.Select(
            options=[],
            # rows=10,
            description='Packages',
            disabled=False,
            layout=Layout(width='90%')
        )

        self.package_dd.observe(self.on_change_package)

        self.dataset_dd = widgets.Select(
            options=[],
            # rows=10,
            description='Datasets',
            disabled=False,
            layout=Layout(width='90%')
        )

        self.dataset_dd.observe(self.on_change_dataset)

        api_keyl = [self.api_key]
        which = [self.which_packages]
        items = [
            widgets.VBox([self.package_dd], layout=Layout(width='50%')),
            widgets.VBox([self.dataset_dd], layout=Layout(width='50%'))
        ]

        tab_contents = ['.display()', '.shape', '.description', 'print', '__iter__']
        children = [widgets.Textarea(
            value='',
            placeholder='',
            disabled=False,
            layout=Layout(width='100%', height='180px')
        ) for name in tab_contents]

        self.tab = widgets.Tab(
            layout=Layout(width='100%', height='250px')
        )

        self.tab.children = children
        [self.tab.set_title(num, name) for num, name in enumerate(tab_contents)]

        vbox3 = widgets.HBox(which, layout=Layout(width='100%', height='50px'))
        vbox2 = widgets.VBox([widgets.HBox(api_keyl),
                              widgets.HBox([self.env_dd, self.button])])
        vbox = widgets.HBox(items)
        vbox4 = widgets.HBox([self.tab])
        self.root = widgets.VBox([vbox2, vbox3, vbox, vbox4, self.copy_button])

    def _generate_name(self, type_):
        while True:
            self._name_counts[type_] = (
                self._name_counts[type_] + 1
            )

            name = '{}_{}'.format(
                type_, self._name_counts[type_]
            )

            if name not in self.ipython.user_global_ns:
                break
            else:
                return self._generate_name(type_)

        return name

    def btn_copyhandler(self, obj):
        _name_alias = {
            'PackageModel': 'package',
            'DatasetModel': 'dataset',
        }

        name = self._generate_name(
            _name_alias[type(self.current_obj).__name__]
        )

        self.ipython.user_global_ns[name] = self.current_obj
        encoded_code = base64.b64encode(name.encode()).decode()
        IPython.display.display(IPython.display.Javascript("""
            var code = IPython.notebook.insert_cell_{0}('code');
            code.set_text(atob("{1}"));
        """.format('below', encoded_code)))


    def btn_eventhandler(self, obj):
        env_map = {
            '1':'https://catalogue.datalake.ihsmarkit.com/__api',
            '2':'https://catalogue-uat.datalake.ihsmarkit.com/__api',
            '3':'https://catalogue-qa.datalake.ihsmarkit.com/__api'
        }

        self.dl = session.start_session(
            self.api_key.value,
            root_url=env_map[str(self.env_dd.value)]
        )
        self.packages = self.dl.packages()
        self.package_dd.options = self.packages
        self.datasets = self.packages[list(self.packages.keys())[0]].datasets()
        self.dataset_dd.options = self.datasets

    def on_change_package(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
     
            self.current_obj = change["new"]
            self.dataset_dd.options = self.current_obj.datasets()
            #current_obj will be overwritten
            self.current_obj = change["new"]
            self.tab.children[0].value = redirect(self.current_obj.display)
            self.tab.children[1].value = str(self.current_obj.shape)
            self.tab.children[2].value = str(self.current_obj.description)
            self.tab.children[3].value = str(self.current_obj.__str__())
            self.tab.children[4].value = "\n".join([str(x) for x in
                                                    self.current_obj.datasets()])

    def on_change_dataset(self, change):
            
        if change['type'] == 'change' and change['name'] == 'value':
            self.current_obj = change["new"]
            self.tab.children[0].value = redirect(self.current_obj.display)
            self.tab.children[1].value = ""
            self.tab.children[2].value = ""
            self.tab.children[3].value = str(self.current_obj.__str__())
            self.tab.children[4].value = "\n".join([str(x) for x in
                                                    self.current_obj.instances.all()])


@IPython.core.magic.magics_class
class DatasetSelectorMagic(IPython.core.magic.Magics):

    @IPython.core.magic.line_magic
    def dataset_selector(self, line):
        disp = SelectorWidget(self.parent)
        self.parent.user_global_ns['_widget'] = disp
        IPython.display.display(disp.root)


def load_ipython_extension(ipython):
    # The `ipython` argument is the currently active `InteractiveShell`
    # instance, which can be used in any way. This allows you to register
    # new magics or aliases, for example.
    ipython.register_magics(DatasetSelectorMagic)

def unload_ipython_extension(ipython):
    # If you want your extension to be unloadable, put that logic here.
    pass


test = {}

# def show():
#     IPython.get_ipython().magic('dataset_selector')
#
#
# if __name__ != '__main__':
#     IPython.get_ipython().extension_manager.load_extension('dli.gui')
#     show()



import json
import os
import random
import string
import time
import unittest
import uuid
import jwt
from unittest.mock import patch

import requests
from urllib import parse

from dli.client import builders, session
from tests import localstack_helper

token_data = {
  "aud": "datalake-catalogue-dev",
  "auth_time": 1513331509,
  "datalake": {
    "accounts": {
      "iboxx": "rw",
      "mrd": "r",
      "datalake-mgmt": "rw"
    },
    "tenant_id": "4e27686c-e72a-44f1-a126-f0d2940617d4",
    "organisation_id": "4e27686c-e72a-44f1-a126-f0d2940617d4",
  },
  "email": "jim@example.com",
  "email_verified": True,
  "exp": 9513335109,
  "iat": 1513331509,
  "iss": "https://hydra-dev.udpmarkit.net",
  "nonce": "5a823264-b7d3-485e-b761-ec1512944814",
  "sub": "user:12345:jim"
}

token = jwt.encode(token_data, 'secret').decode('utf')


class SdkIntegrationTestCase(unittest.TestCase):
    """
    Helper TestCase to test against a local docker container running the datacat api.
    To run these locally, run `docker-compose up` on the root directory
    """

    aws_account_id = "12345"

    @patch("dli.client.dli_client.Session._get_auth_key", lambda self: token)
    @patch("dli.client.dli_client.AnalyticsHandler", lambda self: None)
    def setUp(self):
        self.headers = {
            'Content-type': 'application/json',
            'Authorization': 'Bearer {}'.format(token),
            'Cookie': 'oidc_id_token=' + token
        }
        self.root_url = os.environ.get(
            "DATA_LAKE_INTERFACE_URL", "http://data-lake-interface:8080/__api/")
        self.api_key = self.get_api_key()
        self.client = self.create_session()
        self.s3 = []

        while True:
            try:
                #keep sleeping until we can actually register.
                time.sleep(1)
                self.register_aws_account(self.aws_account_id)
                break
            except:
                continue

        self.s3_client = localstack_helper.get_s3_client()

    def patch_upload_files_to_s3(self, files, location, tr=None, r=None):
        result = []
        for f in files:
            path = location + os.path.basename(f)
            self.s3.append(path)
            result.append({"path": "s3://" + path})
        return result

    def get_api_key(self):
        # url = "%s/generate-my-key" % self.root_url
        # payload = {
        #     "account": "datalake-mgmt",
        #     "rights": "rw",
        #     "expiration": "2030-01-01"
        # }

        # response = requests.post(url, data=json.dumps(payload), headers=self.headers)
        # return response.text
        return "key"

    def create_package(self, name, access="Restricted", **kwargs):
        payload = {
            "name": name + "_" + str(uuid.uuid4()),
            "description":  "asd",
            "publisher":  "datalake-mgmt",
            "manager_id":  "datalake-mgmt",
            "access_manager_id":  "datalake-mgmt",
            "tech_data_ops_id":  "datalake-mgmt",
            "topic":  "Academic/Education",
            "access":  access,
            "internal_data":  "No",
            "data_sensitivity":  "Public",
            "terms_and_conditions":  "Terms and Conditions"
        }
        payload.update(**kwargs)
        return self.client.register_package(**payload).package_id

    def dataset_builder(self, package_id, dataset_name, **kwargs):
        arguments = dict(kwargs)
        data_preview_type = arguments.pop('data_preview_type', 'NONE')
        default_short_code = ''.join(random.choices(
            string.ascii_letters + string.digits, k=10
        ))
        short_code = arguments.pop('short_code', default_short_code)
        return builders.DatasetBuilder(
            package_id=package_id,
            name=dataset_name,
            description='a testing dataset',
            content_type='Pricing',
            data_format='CSV',
            publishing_frequency='Daily',
            taxonomy=[],
            data_preview_type=data_preview_type,
            short_code=short_code,
            **arguments
        )

    def register_s3_dataset(self, package_id, dataset_name, bucket_name, prefix="prefix"):
        self._setup_bucket_and_prefix(bucket_name, prefix)
        return self.client.register_dataset(
            self.dataset_builder(package_id, dataset_name).with_external_s3_storage(
                bucket_name,
                self.aws_account_id,
                prefix
            )
        )

    def _setup_bucket_and_prefix(self, bucket_name, prefix):
        # API checks whether DataLake can access S3 bucket at prefix. If dataset is registered with S3 location
        # we must create a fake one in localstack.
        s3_client = self.s3_client
        s3_client.create_bucket(Bucket=bucket_name)
        s3_client.put_object(
            Bucket=bucket_name,
            # Ensure key ends in a slash to signify it's a directory
            Key=prefix.rstrip('/') + '/',
            Body=""
        )

        def cleanup_bucket():
            for object_to_delete in s3_client.list_objects(Bucket=bucket_name)['Contents']:
                s3_client.delete_objects(Bucket=bucket_name, Delete={"Objects": [{"Key": object_to_delete['Key']}]})
            s3_client.delete_bucket(Bucket=bucket_name)

        self.addCleanup(cleanup_bucket)


    def create_session(self):
        return session.start_session(self.api_key, self.root_url)

    def register_aws_account(self, aws_account_id):
        response = requests.post(
            parse.urljoin(self.root_url, "me/aws-accounts"),
            headers=self.headers,
            data=json.dumps({
                "awsAccountId": str(aws_account_id),
                "awsAccountName": str(aws_account_id),
                "awsRegion": "eu-west-1",
                "accountIds": ["datalake-mgmt"],
                "awsRoleArn": 'arn:aws:iam::000000000001:user/root'
            })
        )

        self.assertEqual(response.status_code, 200)

    def assert_page_count_is_valid_for_paginated_resource_actions(self, func):
        with self.assertRaises(ValueError):
            func(-1)
        with self.assertRaises(ValueError):
            func(0)
        with self.assertRaises(ValueError):
            func("test")


def eventually(assertion, delay=1, retries=5):
    try:
        return assertion()
    except Exception as e:
        if retries <= 1:
            raise e
        time.sleep(delay)
        return eventually(assertion, delay=delay, retries=retries-1)



from dli.client.components import SirenComponent
from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.models.account_model import AccountModel
from urllib.parse import urljoin


class Accounts(SirenComponent):

    def get_account_by_name(self, account_name):
        """
        Retrieves account details given an account name.

        :param str account_name: The name of the account. For example found
                                 on package tech data ops.

        :returns: account object
        """
        response = self.session.get(
            urljoin(
                self._environment.accounts,
                'api/identity/v2/dl_accounts'
            ),
            params={
                'filter[0][value]': account_name,
                'filter[0][operator]': 'eq',
                'filter[0][field]': 'name'
            }
        )

        response_data = response.json()

        if not response_data['data']:
            self.logger.warning(f'Account with name {account_name} not found')
            return None

        data = response_data['data'][0]

        return AccountModel._from_v2_response(data)



from datetime import datetime
from unittest.mock import Mock, PropertyMock
from freezegun import freeze_time
from moto import mock_s3

from dli.aws import create_refreshing_session


@mock_s3
def test_session_refreshes():
    dli_client = Mock()
    # This is how you give an attribute a side effect
    auth_key = PropertyMock(return_value='abc')
    type(dli_client.session).auth_key = auth_key 

    dli_client.session.token_expires_on = datetime(2000, 1, 2, 0, 0, 0)
    session = create_refreshing_session(dli_client.session).resource('s3')

    with freeze_time(datetime(2000, 1, 1, 0, 0, 0)):
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 1

    with freeze_time(datetime(2000, 1, 3, 0, 0, 0)):
        dli_client.session.token_expires_on = datetime(2000, 1, 4, 0, 0, 0)
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 2
        session.meta.client.list_buckets()
        assert auth_key.call_count == 2



from dli.models.dictionary_model import DictionaryModel


class TestDictionary:
    def test_dictionary_with_fields(self, test_client):
        model = DictionaryModel(
            {'attributes': {'fields': []}, 'id': '123'}, test_client
        )

        assert not test_client.session.get.called

    def test_dictionary_fields(self, test_client):
        model = DictionaryModel(
            {'attributes': {}, 'id': '123'}, test_client
        )
        model.fields

        assert test_client.session.get.called

    def test_dictionary_field_pagination(self, test_client):

        def _response():
            i = 0
            while True:
                #as the default page_size is 25, to get
                #'6' we effectively *3 as we have two items
                yield {
                    'meta': {'total_count':75},
                    'data': {
                        'attributes': {
                            'fields': [
                                {f'{i}': i},
                                {f"-{i}": i}
                            ]
                        }
                    }
                }
                i += 1

        test_client._session.get().json.side_effect = _response()
        model = DictionaryModel(
            {'attributes': {}, 'id': '123'}, test_client
        )

        assert len(model.fields) == 6



import inspect


class LoggingAspect:

    @staticmethod
    def invoke_pre_call_aspects(wrapped_object, metadata):
        if getattr(wrapped_object, 'logger', None):
            try:
                wrapped_object.logger.info('Client Function Called', extra=metadata)
            except Exception as e:
                wrapped_object.logger.exception(
                    'Error while invoking pre-call aspects.', e
                )

    @staticmethod
    def invoke_post_call_aspects(wrapped, metadata):
        pass

    @staticmethod
    def invoke_after_exception_aspects(wrapped_object, metadata, exception):
        if getattr(wrapped_object, 'logger', None):
            # If the object contains a `strict` boolean and it is set to
            # True, then print out the exception message and a full stack
            # trace.
            # DEFAULTED to True to match the current behaviour. This default
            # can be changed in a later release. When we change the default
            # we will have to update the tests to explicitly set strict to
            # True to maintain the previous test behaviour.
            if getattr(wrapped_object, 'strict', True):
                try:
                    wrapped_object.logger.exception(
                        'Unhandled Exception', stack_info=False, extra={
                            'locals': inspect.trace()[-1][0].f_locals
                        })
                except Exception as e:
                    wrapped_object.logger.exception(
                        'Error while invoking pre-call aspects.', e
                    )
            else:
                # Data scientists do not want to see stack dumps by default,
                # especially when we have a root cause that triggers secondary
                # exceptions.
                wrapped_object.logger.warning(
                    '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                    '\nAn exception occurred so we are returning None.'
                    '\nTo see the exception and stack trace, please start '
                    'the session again with the parameter `strict=True`'
                    '\n<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<'
                )


class AnalyticsAspect:

    @staticmethod
    def invoke_pre_call_aspects(wrapped_object, metadata):
        pass

    def invoke_post_call_aspects(self, wrapped_object, metadata):
        if getattr(wrapped_object, '_analytics_handler', None):
            try:
                self._create_event(wrapped_object, metadata, status_code=200)
            except Exception as e:
                wrapped_object.logger.exception(
                    'Error while invoking post-call aspects.', e
                )

    def invoke_after_exception_aspects(self, wrapped_object, metadata, exception):
        if getattr(wrapped_object, '_analytics_handler', None):
            try:
                status_code = self._retrieve_status_code_from_exception(exception)
                self._create_event(wrapped_object, metadata, status_code=status_code)
            except Exception as e:
                wrapped_object.logger.exception(
                    'Error while invoking after-exception aspects.', e
                )

    @staticmethod
    def _create_event(dli_client, metadata, status_code):
        additional_properties = metadata.get('properties') or {}
        dli_client._analytics_handler.create_event(
            metadata['subject'], metadata['organisation_id'],
            metadata['func'].__qualname__.split('.')[0],
            metadata['func'].__name__,
            {**metadata['arguments'], **metadata['kwargs'], **additional_properties},
            result_status_code=status_code
        )

    @staticmethod
    def _retrieve_status_code_from_exception(exception):
        try:
            return exception.response.status_code
        except AttributeError:
            return 500


def analytics_decorator(function, class_fields_to_include):
    aspect = AnalyticsAspect()

    def function_wrapper(target, *args, **kwargs):
        metadata = extract_metadata(
            target._client, target, function, args, kwargs, class_fields_to_include
        )

        aspect.invoke_pre_call_aspects(target._client, metadata)
        try:
            result = function(target, *args, **kwargs)
        except Exception as e:
            aspect.invoke_after_exception_aspects(target._client, metadata, e)
            raise e
        aspect.invoke_post_call_aspects(target._client, metadata)
        return result

    return function_wrapper


def logging_decorator(function, class_fields_to_include):
    aspect = LoggingAspect()

    def function_wrapper(target, *args, **kwargs):
        metadata = extract_metadata(
            target._client, target, function, args, kwargs, class_fields_to_include
        )
        aspect.invoke_pre_call_aspects(target._client, metadata)
        try:
            result = function(target, *args, **kwargs)
        except Exception as e:
            aspect.invoke_after_exception_aspects(target._client, metadata, e)
            raise e
        aspect.invoke_post_call_aspects(target._client, metadata)
        return result

    return function_wrapper


def extract_metadata(
        dli_client, wrapped_object, func, arguments,
        keyword_args, class_fields_to_include=None
):
    # Get the user calling the function
    org_id, subject = _retrieve_user_details(dli_client)

    # This is to find out what the 'arg' names are.
    argspec = inspect.getfullargspec(func)
    args_dict = dict(zip(argspec.args, [wrapped_object, *arguments]))

    properties = _read_field_values(class_fields_to_include, wrapped_object)

    if not subject:
        if args_dict.get('api_key'):
            subject = '***' + args_dict.get('api_key')[:6]
        else:
            subject = 'UNKNOWN USER'

    return {
        'func': func,
        'subject': subject,
        'organisation_id': org_id,
        'arguments': args_dict,
        'kwargs': dict(keyword_args),
        'properties': properties
    }


def _retrieve_user_details(dli_client):
    org_id = None
    user_id = getattr(dli_client, 'api_key', '')[:6]
    try:
        org_id = dli_client.session.decoded_token.get('datalake').get('organisation_id')
        user_id = dli_client.session.decoded_token.get('sub', 'UNKNOWN USER')
    except AttributeError:
        pass

    return org_id, user_id


def _read_field_values(class_fields_to_include, wrapped_object):
    properties = {}
    for field in class_fields_to_include or []:
        property_name = field.split('.')[-1]
        property_value = wrapped_object
        for nested_field_name in field.split('.'):
            property_value = property_value.__dict__[nested_field_name]
        properties.update({property_name: property_value})
    return properties



import json
import os
import textwrap
import uuid
import warnings
import pandas
import pyarrow
import boto3

from typing import List
from urllib.parse import urljoin
from tabulate import tabulate
from s3transfer.manager import TransferManager

from dli.aws import create_refreshing_session
from dli.models import log_public_functions_calls_using, SampleData, \
    AttributesDict
from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls, dataset_urls
from dli.client.exceptions import DataframeStreamingException, \
    UnAuthorisedAccessException
from dli.models.dictionary_model import DictionaryModel
from dli.models.file_model import get_or_create_os_path


@log_public_functions_calls_using(
    [analytics_decorator, logging_decorator], class_fields_to_log=['dataset_id']
)
class DatasetModel(AttributesDict):

    @property
    def sample_data(self):
        return SampleData(self)

    @property
    def id(self):
        return self.dataset_id

    def __init__(self, **kwargs):
        super().__init__(**kwargs,)
        self.instances = self._client._InstancesCollection(dataset=self)
        self.fields_metadata = None

    @classmethod
    def _from_v2_response(cls, response_json):
        return cls._construct_dataset_using(
            response_json['data']['attributes'], response_json['data']['id']
        )

    @classmethod
    def _from_v2_response_unsheathed(cls, response_json):
        return cls._construct_dataset_using(
            response_json['attributes'], response_json['id']
        )

    @classmethod
    def _from_v1_response_to_v2(cls, v1_response):
        response = cls._client.session.get(
            dataset_urls.v2_instance.format(
                id=v1_response['properties']['datasetId']
            )
        )

        return cls._from_v2_response(response.json())

    @classmethod
    def _from_v2_list_response(cls, response_json):
        return [
            cls._construct_dataset_using(
                dataset['attributes'], dataset['id']
            )
            for dataset in response_json['data']
        ]

    @classmethod
    def _construct_dataset_using(cls, attributes, dataset_id):
        location = attributes.pop('location')
        # In the interests of not breaking backwards compatability.
        # TODO find a way to migrate this to the new nested API.
        if not location:
            location = None
        else:
            location = location[next(iter(location))]
        return cls(
            **attributes,
            location=location,
            dataset_id=dataset_id
        )

    def download(self, destination_path, flatten=False, filter_path=None):
        """
        Downloads all files from the latest instance into the provided
        destination path.
        This is a short-cut function for:
        `Dataset.instances.latest().download(destination_path)`

        :param destination_path: required. The path on the system, where the
            files should be saved. Must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :return: the list of the downloaded files
        """

        if not self.has_access:
            raise UnAuthorisedAccessException(
                'Unfortunately the user you are using '
                'does not have access to this dataset. '
                'Please request access to the package/dataset '
                "to be able to retrieve this content."
            )

        s3_resource = create_refreshing_session(
            dli_client_session=self._client.session
        ).resource(
            's3',
            endpoint_url=f'https://{self._client._environment.s3_proxy}'
        )

        s3_client = s3_resource.meta.client

        with TransferManager(s3_client) as transfer_manager:

            bucket = s3_resource.Bucket(
                self.organisation_short_code
            )

            _paths = []

            filter_prefix = self.short_code + (
                    '' if self.short_code.endswith('/') else '/'
            )

            if filter_path:
                filter_prefix = filter_prefix + filter_path.lstrip('/')

            # TODO parallelize this again?
            for object_summary in bucket.objects.filter(
                # Prefix searches for exact matches and folders
                Prefix=filter_prefix
            ):
                if not object_summary.key.endswith('/'):
                    to_path = get_or_create_os_path(
                        object_summary.key,
                        to=destination_path,
                        flatten=flatten
                    )

                    print(f'Downloading {object_summary.key} to: {to_path}...')

                    if os.path.exists(to_path):
                        warnings.warn(
                            'File already exists. Overwriting.'
                        )

                    # returns a future
                    transfer_manager.download(
                        self.organisation_short_code,
                        object_summary.key,
                        to_path
                    )

                    _paths.append(to_path)

            return _paths

    def _dataframe(self, nrows=None, partitions: List[str] = None, raise_=True):
        warnings.warn(
            'This method is deprecated. Please use `dataframe` (note the '
            'underscore has been removed)',
            DeprecationWarning
        )
        self.dataframe(nrows=nrows, partitions=partitions, raise_=raise_)

    def dataframe(self, nrows=None, partitions: List[str] = None, raise_=True):
        """
        Return the data from the files in the latest instance of the dataset
        as a pandas DataFrame.

        We currently support .csv and .parquet as our data file formats. The
        data files in the latest instance could all be .csv format or all be
        .parquet format. If there is a mix of .csv and .parquet or some other
        file format then we will not be able to parse the data and will
        return an error message.

        :param int nrows: Optional. The max number of rows to return.
            We use the nrows parameter to limit the amount of rows that are
            returned, otherwise for very large dataset it will take a long time
            or you could run out of RAM on your machine!
            If you want all of the rows, then leave this parameter set to the
            default None.

        :param List[str] partitions: Optional. A dict of filters (partitions) to
            apply to the dataframe request in the form of: ["a=12","b>20190110"]
            - will permit whitespace and equality operators [<=, <, =, >, >=]

        :param bool raise_: Optional. Raise exception if the dataframe stream
            stopped prematurely

        :returns: Pandas DataFrame.
        :rtype: pandas.DataFrame
        """

        if not self.has_access:
            raise UnAuthorisedAccessException(
                'Unfortunately the user you are using '
                'does not have access to this dataset. '
                'Please request access to the package/dataset '
                "to be able to retrieve this content."
            )

        params = {}

        if nrows is not None:
            params['filter[nrows]'] = nrows

        if partitions is not None:
            params['filter[partitions]'] = partitions

        dataframe_url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_dataframe.format(id=self.id)
        )

        # Generate request_id on the SDK side for use in Consumption log
        # messages. This is useful for debugging what seems like a success
        # (no exception) but is not what the user wanted e.g. not the data
        # they expected.
        request_id = str(uuid.uuid1())

        tracing = os.environ.get('DLI_TRACE', '')
        if tracing and tracing.lower() == 'true':
            print(f"DEBUG: Your REQUEST ID is {request_id}")

        response = self._client.session.get(
            dataframe_url, stream=True,
            params=params,
            headers={'X-Request-ID': request_id},
        )

        # Don't decode content. We would like to keep the raw
        # probably gziped steam. Otherwise the stream.read(n) will
        # return a string of length != n.
        response.raw.decode_content = False

        native_file = pyarrow.PythonFile(response.raw, mode='rb')

        # If the response is gzip we need to wrap the
        # encoding in a decompressor.
        if 'Content-Encoding' in response.headers:
            if response.headers['Content-Encoding'] == 'gzip':
                native_file = pyarrow.CompressedInputStream(
                    native_file, 'gzip'
                )

        reader = pyarrow.ipc.open_stream(native_file)
        dataframe = reader.read_pandas()

        # The pyarrow buffered stream reader stops once it
        # reaches the end of the IPC message. Afterwards we
        # get the rest of the data which contains the summary
        # of what we've downloaded including an error message.
        summary_string = native_file.read()
        if summary_string:
            summary = json.loads(summary_string)

            if summary['status'] >= 400:
                exception = DataframeStreamingException(
                    summary, dataframe_url, response=response,
                )

                # Optionally ignore bad data
                if raise_:
                    raise exception
                else:
                    warnings.warn(
                        str(exception),
                        UserWarning
                    )

        return dataframe

    def _partitions(self):
        warnings.warn(
            'This method is deprecated. Please use `partitions` (note the '
            'underscore has been removed)',
            DeprecationWarning
        )
        self.partitions()

    def partitions(self):
        """
        Retrieves the list of available partitions for a given dataset.

        The data onboarding team have structured the file paths on S3 with
        simple partitions e.g. `as_of_date` or `location`.

        Their aim was to separate the data to reduce the size of the
        individual files. For example, data that has a `location` column with
        the options `us`, `eu` and `asia` can be separated into S3 paths like
        so:

            package-name/dataset/as_of_date=2019-09-10/location=eu/filename.csv
            package-name/dataset/as_of_date=2019-09-10/location=us/filename.csv

        in this case the `partitions` will be returned as:

            {'as_of_date': ['2019-09-10'], 'location': ['eu', 'us]}
        """
        response = self._client.session.get(
            urljoin(
                self._client._environment.consumption,
                consumption_urls.consumption_partitions.format(id=self.id)
            )
        )

        return response.json()["data"]["attributes"]["partitions"]

    def contents(self):
        """
        Print IDs for all the instances in this dataset.

        Example output:

            INSTANCE 1111aaa-11aa-11aa-11aa-111111aaaaaa
        """
        for p in self.instances.all():
            print(str(p))

    def dictionary(self) -> List[dict]:

        if self.fields_metadata is None:
            # we have to do two calls
            # to get the latest dictionary id for the dataset
            try:
                response = self._client.session.get(
                    dataset_urls.dictionary_by_dataset_lastest.format(id=self.id)
                ).json()
            except:
                print("There is no current dictionary available.")
                return []

            # followed by the fields for the dictionary...


            self.fields_metadata = DictionaryModel(
                {'id': response["data"]["id"], 'attributes': {}}, client=self._client
            ).fields

            def subdict(field_dict, keep):
                return{k: v for k, v in field_dict.items() if k in keep}

            self.fields_metadata = list(
                map(lambda field: subdict(field, ["name", "nullable", "type"]),
                    self.fields_metadata)
            )

        return self.fields_metadata

    def info(self):
        fields = self.dictionary()

        df = pandas.DataFrame(fields)
        if df.shape[1] > 0:
            df["type"] = df.apply(
                lambda row: row["type"] + (" (Nullable)" if row["nullable"] else ""),
                axis=1)
            df = df[["name", "type"]]

            print(tabulate(df, showindex=False, headers=df.columns))
        else:
            print("No columns/info available.")

    def __repr__(self):
        return f'<Dataset short_code={self.short_code}>'

    def __str__(self):
        separator = "-" * 80
        splits = "\n".join(textwrap.wrap(self.description))

        return f"\nDATASET \"{self.short_code}\" [{self.data_format}]\n" \
               f">> Shortcode: {self.short_code}\n"\
               f">> Available Date Range: {self.first_datafile_at} to {self.last_datafile_at}\n" \
               f">> ID: {self.id}\n" \
               f">> Published: {self.publishing_frequency} by {self.organisation_name}\n" \
               f">> Accessible: {self.has_access}\n" \
               f"\n" \
               f"{splits}\n" \
               f"{separator}"



import types
from functools import wraps

from dli.client.aspects import extract_metadata, LoggingAspect, AnalyticsAspect


class ComponentsAspectWrapper(type):
    """
    This decorates all functions in a Component with a logging function.
    """
    __aspects = [LoggingAspect(), AnalyticsAspect()]

    def __new__(cls, name, bases, attrs):
        for attr_name, attr_value in attrs.items():
            if (
                isinstance(attr_value, types.FunctionType)
                and not attr_name.startswith('_')
            ):
                attrs[attr_name] = cls._wrap_call_with_aspects(attr_value)

        return super(ComponentsAspectWrapper, cls).__new__(cls, name, bases, attrs)

    @classmethod
    def _wrap_call_with_aspects(cls, func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                metadata = extract_metadata(self, self, func, args, kwargs)
            except Exception as e:
                if getattr(self, 'logger', None):
                    self.logger.exception(
                        'Error while reading function metadata.', e
                    )
                return func(self, *args, **kwargs)

            try:
                cls._invoke_pre_call_aspects(self, metadata)
                result = func(self, *args, **kwargs)
                cls._invoke_post_call_aspects(self, metadata)
                return result
            except Exception as e:
                # If the object contains a `strict` boolean and it is set to
                # True, then print out the exception message and a full stack
                # trace.
                # DEFAULTED to True to match the current behaviour. This default
                # can be changed in a later release. When we change the default
                # we will have to update the tests to explicitly set strict to
                # True to maintain the previous test behaviour.
                if getattr(self, 'strict', True):
                    cls._invoke_after_exception_aspects(self, metadata, e)
                    raise e
                else:
                    # Data scientists do not want to see stack dumps by
                    # default, especially when we have a root cause that
                    # triggers secondary exceptions.
                    self.logger.warning(
                        '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                        '\nAn exception occurred so we are returning None.'
                        '\nTo see the exception and stack trace, please start '
                        'the session again with the parameter `strict=True`'
                        '\n<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<'
                    )
                    return None

        return wrapper

    @classmethod
    def _invoke_pre_call_aspects(cls, wrapped_object, metadata):
        for aspect in cls.__aspects:
            aspect.invoke_pre_call_aspects(wrapped_object, metadata)

    @classmethod
    def _invoke_post_call_aspects(cls, wrapped_object, metadata):
        for aspect in cls.__aspects:
            aspect.invoke_post_call_aspects(wrapped_object, metadata)

    @classmethod
    def _invoke_after_exception_aspects(cls, wrapped_object, metadata, exception):
        for aspect in cls.__aspects:
            aspect.invoke_after_exception_aspects(wrapped_object, metadata, exception)


class SirenComponent(metaclass=ComponentsAspectWrapper):

    def __init__(self, client=None):
        self.client = client



import warnings
import urllib

from collections import OrderedDict
from typing import Dict

from dli.models.dataset_model import DatasetModel
from dli.models.paginator import Paginator
from dli.client.components.urls import me_urls, search_urls, package_urls
from dli.models.package_model import PackageModel


class PackageModule:

    def __call__(self, search_term=None, only_mine=True) \
            -> Dict[str, PackageModel]:
        """
        See packages we can access at the top level.

        :param str search_term: The search phrase to search the catalogue
        with, and to pull back packages based on name, description etc.

        :param bool only_mine: Specify whether to collect packages only
        accessible to you (user) or to discover packages that you may
        want to discover.

        :returns: Ordered dictionary of id to PackageModel.
        :rtype: OrderedDict[id: str, PackageModel]
        """

        warnings.warn(
            "Sorry, we have had to disable this endpoint.",
            RuntimeWarning
        )
        return OrderedDict([])

        query = {}
        if search_term:
            query['query'] = search_term

        # First get the search term - if it's empty it will return everything
        search_paginator = Paginator(
            search_urls.search_packages + '?' + urllib.parse.urlencode(query),
            self._client._Package,
            self._client._Package._from_v1_response_to_v2
        )

        package_search = OrderedDict([(v.name, v) for v in search_paginator])

        if only_mine:
            consumed_paginator = Paginator(
                me_urls.consumed_packages,
                self._client._Package,
                self._client._Package._raw_dict
            )

            consumed_package_ids = set(
                package['packageId'] for package in consumed_paginator
            )

            # todo this should change to (v.shortcode, v) in future.
            # Short code is not exposed in the package API
            package_search = OrderedDict([
                (v.name, v) for v in package_search.values() if v.id in consumed_package_ids
            ])

        return package_search

    def _get(self, short_code) -> PackageModel:
        """
        Find a PackageModel with the matching short code. If not found then
        returns None.

        :param str short_code: The short code of the package to collect
        :returns: PackageModel with matching short code.
        :rtype: PackageModel
        """
        warnings.warn(
            'Getting a dataset by name, and package name or package ID '
            'will be deprecated in future. Short-codes will replace this.',
            PendingDeprecationWarning
        )
        params = {}

        # todo this should be by short code but EP doesnt seem to exist for this yet
        # so pass p_id
        params["name"] = short_code
        response = self._client.session.get(
            package_urls.package_index, params=params
        )
        return self._client._Package._from_v1_response_to_v2(response.json())

    def get_dataset(self, package_name, dataset_short_code) -> DatasetModel:
        """
        Find a PackageModel and DatasetModel with the matching short code
        within the package. If not found then returns None.

        :param str package_name: required. The name of the package
            in which we search for a dataset.
        :param str dataset_short_code: required. The short code of the dataset
            to collect from within the package.
        :return: DatasetModel
        """
        if package_name is None:
            raise ValueError('package_name must not be None')

        if dataset_short_code is None:
            raise ValueError('dataset_short_code must not be None')

        params = {'name': package_name}
        package = self._client.session.get(
            package_urls.package_index, params=params
        ).json()

        params = {
            'filter[0][field]': 'short_code',
            'filter[0][operator]': 'eq',
            'filter[0][value]': dataset_short_code
        }

        dataset = self._client._session.get(
            package_urls.v2_package_datasets.format(
                id=package['properties']['packageId']),
            params=params
        )

        return dataset.json()



try:
    import simplejson as json  # noqa: I900
except ImportError:
    import json


class DatalakeException(Exception):
    def __init__(self, entity=None, params=None, message=None, response=None):
        if not params and not message:
            message = "%s not found" % entity

        self.params = params
        self.message = (
            message or 'entity with params `{}` not found'.format(params)
        )
        self.entity = entity
        self.response = response

        super(DatalakeException, self).__init__(
            self.message, entity,
            params, message
        )

    def __str__(self):
        error = f'\n{self.response.request.method} {self.response.request.url}\n\n'

        if self.response.status_code == 413:
            error += (
                'A file you are attempting to pass is probably too large for our server.'
                ' If passing a sample_data file please make sure it is under 10 Megabytes.\n\n'
            )

        try:
            error += (
                f'{json.dumps(self.response.json(), indent=4)}'
            )
        except json.JSONDecodeError:
            error += (
                f'{self.response.text}'
            )

        return error


class DataframeStreamingException(DatalakeException):

    def __init__(self, error, url, entity=None, params=None, message=None,
                 response=None):
        self.url = url
        self.error = error
        super().__init__(
            entity=entity,
            params=params,
            message=message,
            response=response
        )

    def __str__(self):
        msg = json.dumps(self.error, indent=4)
        return f'\nGET {self.url}\n\n{msg}'


class CatalogueEntityNotFoundException(DatalakeException):
    pass


class InvalidPayloadException(DatalakeException):
    pass


class UnAuthorisedAccessException(DatalakeException):
    pass


class InsufficientPrivilegesException(DatalakeException):
    message = 'Insufficient privileges to perform this action'


class S3FileDoesNotExist(DatalakeException):
    def __init__(self, file_path):
        self.message = (
            "Either file at path `%s` does not exist / Potential issue with the bucket policy."
            "Please reach out to Datalake Tech Data Ops user for resolution." % file_path
        )

        super(S3FileDoesNotExist, self).__init__(self.message)


class DownloadFailed(DatalakeException):
    pass


class NoAccountSpecified(DatalakeException):

    def __init__(self, accounts):
        self.accounts = accounts
        self.message = (
            "Unable to default the account for access_manager_id, tech_data_ops_id and/or manager_id "
            "due to multiple accounts being attached to this API key. "
            "Your accounts are: %s" % [(a.id, a.name) for a in accounts]
        )
        super(NoAccountSpecified, self).__init__(self.message)


class AuthenticationFailure(Exception):
    """
    An exception wrapping an authentication failure response. If the response
    had a payload, that payload is reported as the exception message, otherwise
    a generic error message is returned.
    """
    GENERIC_ERROR_MESSAGE = (
        'API key is not valid or has expired.'
        ' Please generate a new key in the Catalogue UI and try again.'
    )

    def __init__(self, response=None, message=None):
        self.response = response
        self.message = message

    def __str__(self):
        if self.response and self.response.text:
            return self.response.text
        elif self.message:
            return self.message
        return self.GENERIC_ERROR_MESSAGE



import textwrap
import warnings
import urllib
from collections import OrderedDict
from typing import Dict

from dli.models.paginator import Paginator
from dli.client.components.urls import search_urls, dataset_urls
from dli.models.dataset_model import DatasetModel
from dli.client.exceptions import CatalogueEntityNotFoundException


class DatasetModule:

    def __call__(self, search_term=None, only_mine=True) \
            -> Dict[str, DatasetModel]:
        """
        See datasets.

        :param str search_term: The search phrase to search the catalogue
        with, and to pull back datasets based on name, description etc.

        :param bool only_mine: Specify whether to collect datasets only
        accessible to you (True) or to discover packages that you may
        want to discover but may not have access to (False).

        If any combination of the two, then search term takes precedence
        of routing, followed by the application of only_mine flag.

        :returns: Ordered dictionary of ids to DatasetModel.
        :rtype: OrderedDict[id: str, DatasetModel]
        """

        warnings.warn(
            "Sorry, we have had to disable this endpoint.",
            RuntimeWarning
        )
        return OrderedDict([])

        query = {}
        if search_term:
            query['query'] = search_term

        p = Paginator(
            # Maybe refactor out into a DatasetPaginator
            search_urls.search_datasets + '?' + urllib.parse.urlencode(query),
            self._client.Dataset,
            self._client.Dataset._from_v1_response_to_v2,
            max_workers=5
        )

        warning_message = \
            textwrap.dedent("""
            Warning! There are duplicate short codes for
            datasets. The short code in the dictionary of dataset we are now 
            returning will refer to only one of the datasets. 
            If you are running in the production or UAT environments then
            please report the problem to the datalake 
            platform team at DataLake-Support_L1@ihsmarkit.com. 
            Please include this full warning message:
            """)

        if only_mine:
            warnings.warn('This search may return datasets which exist but '
                          'are not accessible to your user. You may have to '
                          'request access.')
            len_before_filtering = len([v for v in p])
            count_has_access_unknown = len(
                [v for v in p if not isinstance(v.has_access, bool)])
            count_has_access_false = len([
                v for v in p if
                isinstance(v.has_access, bool) and not v.has_access
            ])
            count_has_access_true = len([
                v for v in p if
                isinstance(v.has_access, bool) and v.has_access
            ])

            o = OrderedDict([
                # Hack until better filtering
                (v.short_code, v) for v in p
                if isinstance(v.has_access, bool) and v.has_access
            ])

            len_after_dedup = len(o)

            if count_has_access_true != len_after_dedup:
                warnings.warn(
                    textwrap.dedent(f"""{warning_message}
                    \nLength before filtering '{len_before_filtering}',
                    \ncount has access unknown '{count_has_access_unknown}',
                    \ncount has access false '{count_has_access_false}',
                    \ncount has access true '{count_has_access_true}',
                    \nlength of de-duplicated dict '{len_after_dedup}',
                    \ndifference after de-duplicating 
                    '{count_has_access_true - len_after_dedup}'"""),
                    RuntimeWarning
                )

            return o
        else:
            len_before_dedup = len([v for v in p])

            o = OrderedDict([
                (v.short_code, v) for v in p
            ])

            len_after_dedup = len(o)

            if len_before_dedup != len_after_dedup:
                warnings.warn(
                    textwrap.dedent(f"""{warning_message}
                    \nlength before de-duplicating '{len_before_dedup}',
                    \nlength after de-duplicating '{len_after_dedup}',
                    \ndifference after de-duplicating
                    '{len_before_dedup - len_after_dedup}'"""),
                    RuntimeWarning
                )

            return o

    def _get(self, short_code) -> DatasetModel:
        """
        Returns a DatasetModel if it exists else None

        :param str short_code: The short code of the dataset to collect

        :returns: Dataset model with matching short code.
        :rtype: DatasetModel
        """

        return self._client.get_dataset(dataset_short_code=short_code)



import pytest
import freezegun
import datetime

@pytest.mark.integration
def test_client_refresh(client, empty_dataset):
    client.get_dataset(empty_dataset.id)
    
    with freezegun.freeze_time(
        client.session.token_expires_on + datetime.timedelta(days=10)
    ):
        client.get_dataset(empty_dataset.id)



from urllib.parse import urljoin

import dli
from dli.client.components.urls import consumption_urls


class AnalyticsHandler:
    __acceptable_properties = [
        'package_id', 'dataset_id', 'name', 'dataset_name', 'datafile_id',
        'dictionary_id', 'api_key'
    ]

    def __init__(self, dli_client):
        self._app_name = 'SDK'
        self._app_version = dli.__version__
        self._client = dli_client
        self._logger = dli_client.logger

    def create_event(
        self, user_id, organisation_id,
        entity=None, action=None, properties=None, result_status_code=None
    ):
        self._log_call_handling_exceptions(
            user_id, organisation_id, entity, action, properties,
            result_status_code
        )

    def _log_call_handling_exceptions(
        self, user_id, organisation_id, entity, action, properties, result
    ):
        try:
            body = self._prepare_body_to_send(
                user_id, organisation_id, entity, action, properties, result,
            )
            self._send_event_to_analytics(body)
        except Exception as e:
            # Data scientists do not want to see stack dumps by default,
            # especially when we have a root cause that triggers secondary
            # exceptions.
            if self._client.strict:
                self._logger.exception('Error while sending analytics: ', e)

    def _prepare_body_to_send(
        self, user_id, organisation_id, entity, action, properties, result
    ):
        properties_to_send = self._filter_out_properties(properties)
        properties_to_send = self._override_properties_for_dataset_functions(
            action, entity, properties, properties_to_send)
        if 'api_key' in properties:
            properties['api_key'] = properties['api_key'][:6]
        event = {
            'application_name': self._app_name,
            'application_version': self._app_version,
            'user_id': user_id,
            'entity': entity,
            'action': action,
            'organisation_id': organisation_id,
            'result': result,
            'properties': properties_to_send
        }
        body = {'data': [{'attributes': event}]}
        return body

    def _filter_out_properties(self, properties):
        return {
            k: v
            for k, v in properties.items() if k in self.__acceptable_properties
        }

    # This is ugly, but 1) register_dataset accepts builder,
    # so we need to extract data from it 2) some functions accept 'id'
    # instead of dataset_id. We either ovveride it like this or not make it
    # generic at all.
    def _override_properties_for_dataset_functions(
            self, action, entity, properties, props
    ):
        if entity == 'Dataset':
            if action == 'register_dataset':
                props = self._filter_out_properties(properties['builder']._data)
            elif 'id' in properties:
                props['dataset_id'] = properties['id']
        return props

    def _send_event_to_analytics(self, body):
        url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_analytics
        )
        session = self._client.session
        session.post(url, json=body)



from unittest import mock

import pytest

from unittest.mock import Mock, MagicMock, call

from dli.client.components import ComponentsAspectWrapper
from dli.client.dli_client import DliClient


class Client(DliClient):

    def __init__(self):
        self.logger = MagicMock()
        session = Mock(
            decoded_token={'sub': 'sub', 'datalake': {'organisation_id': 'tid'}}
        )
        self._session = session

    @property
    def session(self):
        return self._session

    def test_function(self, attr1, attr2, attr3=None):
        pass

    def test_exception_function(self, attr1, attr2, attr3=None):
        raise ValueError('abc')


@pytest.fixture
def aspects():
    yield [Mock(), Mock()]


@pytest.fixture
def client(aspects):
    with mock.patch.object(
        ComponentsAspectWrapper, '_ComponentsAspectWrapper__aspects', aspects
    ):
        yield Client()


@pytest.fixture
def metadata(client):
    return {
        'func': Any(),
        'subject': 'sub',
        'organisation_id': 'tid',
        'arguments': {'self': client, 'attr1': 'val1', 'attr2': 'val2'},
        'kwargs': {'attr3': 'val3'},
        'properties': {}
    }


def Any():
    class Any:
        def __eq__(self, other):
            return True
    return Any()


class TestComponentsAspectWrapper:

    def test_calls_aspects_when_no_exception(self, client, aspects, metadata):
        expected_call = call(client, metadata)

        client.test_function('val1', 'val2', attr3='val3')

        # one call is for __init__ method, the other for the actual test_func
        assert aspects[0].invoke_pre_call_aspects.call_count == 1
        # assert aspects[0].invoke_pre_call_aspects.call_args_list[1] == expected_call
        assert aspects[0].invoke_post_call_aspects.call_count == 1
        # assert aspects[0].invoke_post_call_aspects.call_args_list[1] == expected_call
        assert aspects[0].invoke_after_exception_aspects.call_count == 0

        assert aspects[1].invoke_pre_call_aspects.call_count == 1
        # assert aspects[1].invoke_pre_call_aspects.call_args_list[1] == expected_call
        assert aspects[1].invoke_post_call_aspects.call_count == 1
        # assert aspects[1].invoke_post_call_aspects.call_args_list[1] == expected_call
        assert aspects[1].invoke_after_exception_aspects.call_count == 0

    def test_calls_aspects_when_exception_occurs(
            self, client, aspects, metadata
    ):
        expected_call_no_exception = call(client, metadata)

        with pytest.raises(ValueError) as e:
            client.test_exception_function('val1', 'val2', attr3='val3')


        expected_call_exception = call(client, metadata, e.value)

        assert aspects[0].invoke_pre_call_aspects.call_count == 1
        assert aspects[0].invoke_pre_call_aspects.call_args_list[0] == \
               expected_call_no_exception

        assert aspects[0].invoke_post_call_aspects.call_count == 0
        assert aspects[0].invoke_after_exception_aspects.call_count == 1
        assert aspects[0].invoke_after_exception_aspects.call_args_list[0] == \
               expected_call_exception



import pytest

from unittest.mock import MagicMock, Mock

from dli.client.components import AnalyticsAspect


@pytest.fixture
def analytics_handler():
    yield Mock()


@pytest.fixture
def wrapped_object(analytics_handler):
    obj = Mock()
    obj._analytics_handler = analytics_handler
    yield obj


@pytest.fixture
def metadata():
    return {
        'subject': 'sub', 'organisation_id': 'org_id',
        'func': MagicMock(__qualname__='Dataset.get', __name__='get'),
        'arguments': {'why_not': 42}, 'kwargs': {}
    }


class TestLoggingAspect:

    analytics_aspect = AnalyticsAspect()

    def test_does_not_create_event_pre_call_if_handler_exists(
        self, wrapped_object
    ):
        self.analytics_aspect.invoke_pre_call_aspects(wrapped_object, {})

        assert wrapped_object._analytics_handler.create_event.call_count == 0

    def test_does_not_create_event_pre_call_if_handler_not_defined(
        self, wrapped_object
    ):
        wrapped_object._analytics_handler = MagicMock(
            __bool__=Mock(return_value=False)
        )
        self.analytics_aspect.invoke_pre_call_aspects(wrapped_object, {})

        assert wrapped_object._analytics_handler.create_event.call_count == 0

    def test_creates_event_post_call_if_handler_exists(
        self, wrapped_object, metadata
    ):
        self.analytics_aspect.invoke_post_call_aspects(wrapped_object, metadata)

        wrapped_object._analytics_handler.create_event\
            .assert_called_once_with(
                'sub', 'org_id', 'Dataset', 'get', {'why_not': 42},
                result_status_code=200
            )

    def test_does_not_create_event_post_call_if_handler_not_defined(
        self, wrapped_object
    ):
        wrapped_object._analytics_handler = MagicMock(
            __bool__=Mock(return_value=False)
        )

        self.analytics_aspect.invoke_post_call_aspects(
            wrapped_object, {}
        )

        assert wrapped_object._analytics_handler.create_event.call_count == 0

    def test_creates_event_after_specific_exception_if_handler_exists(
            self, wrapped_object, metadata
    ):
        exception = Mock()
        exception.response.status_code = 404
        self.analytics_aspect.invoke_after_exception_aspects(
            wrapped_object, metadata, exception
        )

        wrapped_object._analytics_handler.create_event\
            .assert_called_once_with(
                'sub', 'org_id', 'Dataset', 'get', {'why_not': 42},
                result_status_code=404
            )

    def test_creates_event_after_unknown_exception_if_handler_exists(
            self, wrapped_object, metadata
    ):
        exception = Mock()
        exception.response = None
        self.analytics_aspect.invoke_after_exception_aspects(
            wrapped_object, metadata, exception
        )

        wrapped_object._analytics_handler.create_event\
            .assert_called_once_with(
                'sub', 'org_id', 'Dataset', 'get', {'why_not': 42},
                result_status_code=500
            )

    def test_does_not_create_event_after_exception_if_handler_not_defined(
            self, wrapped_object
    ):
        wrapped_object._analytics_handler = MagicMock(
            __bool__=Mock(return_value=False)
        )

        self.analytics_aspect.invoke_after_exception_aspects(
            wrapped_object, {}, None
        )

        assert wrapped_object._analytics_handler.create_event.call_count == 0



import pytest

from unittest.mock import MagicMock
from dli.client.dli_client import DliClient
from dli.client.exceptions import CatalogueEntityNotFoundException


class Client(DliClient):
    def __init__(self):
        self._session = MagicMock()
        self._environment = MagicMock()
        self._environment.accounts = 'https://local.test/'
        self.logger = MagicMock()

    @property
    def session(self):
        return self._session


@pytest.fixture
def client():
    yield Client()


class TestAccounts:
    def test_get_account(self, client):
        client.session.get().json.return_value = {
          'data': [
            {
              'attributes': {
                'tenant_id': '9516c0ba-ba7e-11e9-8b34-000c6c0a981f',
                'organisation_id': '9516c0ba-ba7e-11e9-8b34-000c6c0a981f',
                'data': {
                  'contacts': [
                    {
                      'name': 'Michael Salerno',
                      'role': 'Global Head of Rights Management',
                      'type': 'Business',
                      'email': 'CARMDataLakeChecks@ihsmarkit.com'
                    }
                  ],
                  'division': 'Other',
                  'cost_code': 'N/A',
                  'description': 'Content acquisition and rights management team',
                  'is_internal': True
                },
                'name': 'CARM'
              },
              'id': 'test'
            }
          ],
        }

        account = client.get_account_by_name('CARM')
        assert account.id == 'test'

    def test_get_account_no_results(self, client):
        account_name = 'CARM'
        client.session.get().json.return_value = {'data': []}
        assert client.get_account_by_name(account_name) is None
        client.logger.warning.assert_called_once_with(
            f'Account with name {account_name} not found'
        )



from dli.client.components import SirenComponent
from dli.client.utils import ensure_count_is_valid
from dli.client.components.urls import me_urls


class Me(SirenComponent):
    """Functions related to the current logged in user."""

    def _get_my_entities(self, entity, my_entities_func, count):
        ensure_count_is_valid(count)

        result = my_entities_func(page_size=count)
        return result.get_entities(rel=entity)

    def get_my_packages(self, count=100):
        """
        Returns a list of packages where session user account is: a Manager, Tech Data Ops or Access Manager.

        :param int count: The number of items to retrieve, optional. Defaults to 100.

        :returns: List of my packages.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_packages = client.get_my_packages()

        """
        ensure_count_is_valid(count)

        response = self.session.get(
            me_urls.my_packages, params={'page_size': count}
        )

        return response.to_many_siren('package')

    def get_my_consumed_packages(self, count=100):
        """
        Returns a list of all the packages user session account has access to.

        :param int count: The number of items to retrieve, optional. Defaults to 100.

        :returns: List of my consumed packages.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_consumed_packages = client.get_my_consumed_packages()

        """
        ensure_count_is_valid(count)

        response = self.session.get(
            me_urls.consumed_packages, params={'page_size': count}
        )

        return response.to_many_siren('access_request')

    def get_my_accounts(self):
        """
        Returns a list of all the accounts associated with user session.

        :returns: list of all associated accounts.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_accounts = client.get_my_accounts()

        """
        response = self.session.get(me_urls.my_accounts)
        return response.to_many_siren('')



from urllib.parse import urlparse, ParseResult


class _Environment:


    _catalogue_accounts_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'catalogue.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'catalogue-uat.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'client-uat.datalake.ihsmarkit.com',
        'catalogue-dev.udpmarkit.net': 'catalogue-dev.udpmarkit.net',
        'catalogue-qa.udpmarkit.net': 'catalogue-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 'catalogue-qa2.udpmarkit.net',
    }

    _catalogue_consumption_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'consumption.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'consumption-uat.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'consumption-uat2.datalake.ihsmarkit.com',
        'catalogue-dev.udpmarkit.net': 'consumption-dev.udpmarkit.net',
        'catalogue-qa.udpmarkit.net': 'consumption-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 'consumption-qa2.udpmarkit.net',
    }

    _catalogue_sam_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 'sam.samexternal.net',
        'catalogue-qa2.udpmarkit.net': 'sam.samexternal.net',
        'catalogue-dev.udpmarkit.net': 'sam.samexternal.net',
    }

    _catalogue_sam_client_map = {
        'catalogue.datalake.ihsmarkit.com': 'datalake-pkce-prod-q56PXZSTDQ',
        'catalogue-uat.datalake.ihsmarkit.com': 'datalake-pkce-uat-RMrWmjPhql',
        'client-uat.datalake.ihsmarkit.com': 'datalake-pkce-uat-RMrWmjPhql',
        'catalogue-qa.udpmarkit.net': 'datalake-pkce-sqa-BSFwJUigI4',
        'catalogue-qa2.udpmarkit.net': 'datalake-pkce-sqa-BSFwJUigI4',
        'catalogue-dev.udpmarkit.net': 'datalake-pkce-dev-sS8hXQiT2m',
    }

    _s3_proxy_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 's3.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 's3-uat.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 's3-uat2.datalake.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 's3-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 's3-qa2.udpmarkit.net',
        'catalogue-dev.udpmarkit.net': 's3-dev.udpmarkit.net',
    }

    def __init__(self, api_root):
        """
        Class to manage the different endpoints

        :param str root_url: The root url of the catalogue
        """

        catalogue_parse_result = urlparse(api_root)

        self.catalogue = ParseResult(
            catalogue_parse_result.scheme, catalogue_parse_result.netloc,
            '', '', '', ''
        ).geturl()

        accounts_host = self._catalogue_accounts_environment_map.get(
            catalogue_parse_result.netloc
        )

        self.accounts = ParseResult(
            catalogue_parse_result.scheme, accounts_host, '', '', '', ''
        ).geturl()

        consumption_host = self._catalogue_consumption_environment_map.get(
            catalogue_parse_result.netloc
        )

        self.consumption = ParseResult(
            'https', consumption_host, '', '', '', ''
        ).geturl()

        sam_host = self._catalogue_sam_environment_map.get(
            catalogue_parse_result.netloc
        )

        self.sam = ParseResult(
            'https', sam_host, '', '', '', ''
        ).geturl()

        self.sam_client = self._catalogue_sam_client_map.get(
            catalogue_parse_result.netloc
        )

        self.s3_proxy = self._s3_proxy_environment_map.get(
            catalogue_parse_result.netloc
        )



import pytest
import jwt

from http import HTTPStatus
from unittest import TestCase
from unittest.mock import MagicMock, patch

from dli.client.exceptions import AuthenticationFailure, DatalakeException
from dli.client.dli_client import Session

valid_token = jwt.encode({"exp": 9999999999}, 'secret')

class TestAuth(TestCase):

    @pytest.fixture(autouse=True)
    def session(self):
        environ = MagicMock(catalogue='http://test.local/', accounts='')
        self.session = Session(None, 'api_key', environ, None, valid_token)

    def test_auth(self):
        self.session.send = MagicMock()
        self.session._get_auth_key()
        prep_request = self.session.send.call_args[0][0]
        assert prep_request.headers['Authorization'] == (
            'Bearer api_key'
        )

    def test_handle_auth_error(self):
        self.session.post = MagicMock()
        self.session.post.side_effect = MagicMock(
            side_effect=DatalakeException
        )

        with self.assertRaises(AuthenticationFailure) as cm:
            self.session._get_auth_key()



import warnings

from dli.models.paginator import Paginator
from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import dataset_urls
from dli.models import log_public_functions_calls_using, AttributesDict


@log_public_functions_calls_using(
    [analytics_decorator, logging_decorator], class_fields_to_log=['dictionary_id']
)
class DictionaryModel(AttributesDict):

    @property
    def id(self):
        return self.dictionary_id

    @property
    def schema_id(self):
        """This method is deprecated. Please use DictionaryModel.id"""
        warnings.warn(
            'This method is deprecated. Please use DictionaryModel.id',
            DeprecationWarning
        )
        return self.id

    @property
    def fields(self):

        if 'fields_read' not in self.__dict__:
            self.__dict__['fields'] = list(self._paginator)
            self.__dict__['fields_read'] = True

        return self.__dict__['fields']

    def __init__(self, json, client=None):
        self._client = client

        super().__init__(
            json['attributes'],
            dictionary_id=json['id'],
        )

        def append_flattened_to_cache(x, cache):
            for v in x["attributes"]["fields"]:
                cache.append(Field(v))


        self._paginator =Paginator(
            dataset_urls.dictionary_fields.format(id=self.dictionary_id),
            self._client._DictionaryV2,
            instantiation_override=append_flattened_to_cache,
            total_field="total_count",
            page_size=25
        )


class Field(AttributesDict):
    """
    Represents a dictionary Field. Exists
    to provide a basic class with a name.
    """


import pytest

from unittest.mock import MagicMock, Mock

from dli.client.components import LoggingAspect


@pytest.fixture
def logger():
    yield Mock()


@pytest.fixture
def wrapped_object(logger):
    obj = Mock()
    obj.logger = logger
    yield obj


@pytest.fixture
def metadata():
    yield Mock()


class TestLoggingAspect:

    logging_aspect = LoggingAspect()

    def test_logs_info_pre_call_if_logger_exists(
        self, wrapped_object, metadata
    ):
        self.logging_aspect.invoke_pre_call_aspects(wrapped_object, metadata)

        wrapped_object.logger.info.assert_called_once_with(
            'Client Function Called', extra=metadata
        )
        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0
        assert wrapped_object.logger.exception.call_count == 0

    def test_does_not_log_anything_pre_call_if_logger_not_defined(
        self, wrapped_object, metadata
    ):
        wrapped_object.logger = MagicMock(__bool__=Mock(return_value=False))
        self.logging_aspect.invoke_pre_call_aspects(wrapped_object, metadata)

        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.info.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0
        assert wrapped_object.logger.exception.call_count == 0

    def test_does_not_log_anything_post_call_if_logger_exists(
        self, wrapped_object, metadata
    ):
        self.logging_aspect.invoke_post_call_aspects(wrapped_object, metadata)

        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.info.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0
        assert wrapped_object.logger.exception.call_count == 0

    def test_does_not_log_anything_post_call_if_logger_not_defined(
        self, wrapped_object, metadata
    ):
        wrapped_object.logger = MagicMock(__bool__=Mock(return_value=False))
        self.logging_aspect.invoke_pre_call_aspects(wrapped_object, metadata)

        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.info.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0
        assert wrapped_object.logger.exception.call_count == 0

    def test_logs_exception_after_exception_post_call_if_logger_exists(
            self, wrapped_object, metadata, monkeypatch
    ):
        inspect_mock = Mock(trace=Mock(return_value=[[Mock(f_locals='abc')]]))
        monkeypatch.setattr('dli.client.aspects.inspect', inspect_mock)

        self.logging_aspect.invoke_after_exception_aspects(
            wrapped_object, metadata, None
        )

        wrapped_object.logger.exception.assert_called_once_with(
            'Unhandled Exception', stack_info=False, extra={
                    'locals': 'abc'
                }
        )
        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.info.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0

    def test_does_not_log_anything_after_exception_if_logger_not_defined(
            self, wrapped_object, metadata
    ):
        wrapped_object.logger = MagicMock(__bool__=Mock(return_value=False))
        self.logging_aspect.invoke_after_exception_aspects(
            wrapped_object, metadata, None
        )

        assert wrapped_object.logger.warning.call_count == 0
        assert wrapped_object.logger.info.call_count == 0
        assert wrapped_object.logger.debug.call_count == 0
        assert wrapped_object.logger.error.call_count == 0
        assert wrapped_object.logger.exception.call_count == 0



import datetime
import logging
import posixpath
import warnings

from dli.client.components.urls import datafile_urls
from dli.client.components import SirenComponent
from dli.client.exceptions import DownloadFailed
from dli.client.s3 import Client, S3DatafileWrapper
from dli.siren import siren_to_dict, siren_to_entity

logger = logging.getLogger(__name__)


class Datafile(SirenComponent):

    def register_datafile_metadata(
        self,
        dataset_id,
        name,
        files,
        data_as_of,
        original_name=None
    ):
        """
        Submit a request to create a new datafile under a specified dataset in the Data Catalogue. This function
        ``WILL NOT`` upload files

        Datafile is an instance of the data within a dataset, representing a snapshot of the data at the time of
        publishing. A dataset can be composed by one or more related files that share a single schema. of related
        datafiles. It provides user with metadata required to consume and use the data.

        See description for each parameter, and whether they are optional or mandatory.

        :param str dataset_id: Dataset ID to which the datafile belongs to.
        :param str name: A descriptive name of a datafile. It should be unique within the dataset.
        :param list[dict] files: List of file dictionaries. A file dictionary will contain the full file path
                                and size (optional) as items.
        :param str data_as_of: The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.
        :param str original_name: Optional. Name of the data uploaded into the data lake.

        :returns: Registered datafile
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.register_datafile_metadata(
                    dataset_id,
                    name="My Datafile",
                    files=[{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}],
                    data_as_of="2018-05-28"
                )
        """
        dataset = self.get_dataset(id=dataset_id)

        if not files:
            raise Exception("No files to register have been provided.")

        fields = {
            'datasetId': dataset.dataset_id,
            'name': name,
            'originalName': original_name,
            'dataAsOf': self._ensure_iso_date(data_as_of),
            'files': files,
        }

        payload = {k: v for k, v in fields.items() if v is not None}

        return siren_to_entity(
            self.session.post(
                datafile_urls.datafiles_index, json=payload
            ).to_siren()
        )

    def register_s3_datafile(
        self,
        dataset_id,
        name,
        files,
        s3_prefix,
        data_as_of,
        original_name=None
    ):
        """
        Submit a request to create a new datafile under a specified dataset in the Data Catalogue.
        This function will perform an upload of the files to S3 data store.

        Datafile is an instance of the data within a dataset, representing a snapshot of the data at the time of publishing.
        A dataset can be composed by one or more related files that share a single schema. of related datafiles.
        It provides user with metadata required to consume and use the data.

        See description for each parameter, and whether they are optional or mandatory.

        :param str dataset_id: Dataset ID to which the datafile belongs to.
        :param str name: A descriptive name of a datafile. It should be unique within the dataset.
        :param list[str] files: Path of the files or folders to register.
        :param str s3_prefix: location of the files in the destination bucket. Cannot end with slash ("/")
        :param str data_as_of: The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.
        :param str original_name: Optional. Name of the data uploaded into the data lake.

        :returns: Registered datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.register_s3_datafile(
                    dataset_id=dataset_id,
                    name="My datafile",
                    files=["./test_sandbox/samples/data/AAPL.csv", "./test_sandbox/samples/data/MSFT.csv"],
                    s3_prefix="quotes/20180518",
                    data_as_of="2018-05-28"
                )
        """
        dataset = self.get_dataset(id=dataset_id)

        if not files:
            raise Exception("No files to register have been provided.")

        # upload files
        if dataset.location.type != 'S3':
            raise Exception(
                "Only datasets backed on S3 are supported. use register_datafile_metadata instead.")

        bucket = dataset.location.bucket
        if not bucket:
            raise Exception(
                "Dataset location is S3, however, "
                "there is no bucket associated with the dataset {}".format(
                    dataset_id)
            )

        s3_location = "{}/{}".format(bucket, s3_prefix)
        uploaded = self._process_s3_upload(files, s3_location, dataset_id)
        # register metadata
        return self.register_datafile_metadata(
            dataset_id,
            name=name,
            files=uploaded,
            original_name=original_name,
            data_as_of=self._ensure_iso_date(data_as_of),
        )

    def edit_datafile_metadata(
        self,
        datafile_id,
        name=None,
        original_name=None,
        data_as_of=None,
        files=None
    ):
        """
        Edits metadata for an existing datafile.
        This function WILL NOT upload files.
        Fields passed as ``None`` will retain their original value.

        :param str datafile_id: The id of the datafile we want to modify.
        :param str name: Optional. Name of the datafile.
        :param list[dict] files: Optional. List of file dicts. A file dict will contain file path and size(optional) as items.
        :param str original_name: Optional. Original Name for the datafile.
        :param str data_as_of: Optional. The effective date for the data within the datafile.
                               Expected format is YYYY-MM-DD.

        :returns: Registered datafile
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                updated_datafile = client.edit_datafile_metadata(
                    datafile_id,
                    name="My Datafile",
                    files=[{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
                )
        """

        datafile = self._get_datafile(datafile_id)

        fields = {
            'datasetId': datafile.datasetId,
            'name': name,
            'originalName': original_name,
            'dataAsOf': self._ensure_iso_date(data_as_of),
            'files': files
        }

        # clean up any unknown fields, and update the entity
        datafile_as_dict = siren_to_dict(datafile)
        for key in list(datafile_as_dict.keys()):
            if key not in fields:
                del datafile_as_dict[key]

        payload = {k: v for k, v in fields.items() if v is not None}
        datafile_as_dict.update(payload)

        # perform the update and return the resulting entity
        response = self.session.put(
            datafile_urls.datafiles_instance.format(id=datafile_id),
            json=datafile_as_dict
        )

        return siren_to_entity(response.to_siren())

    def delete_datafile(self, datafile_id):
        """
        Marks a datafile as deleted.

        :param str datafile_id: the unique id for the datafile we want to delete.

        :returns: None

        - **Sample**

        .. code-block:: python

                client.delete_datafile(datafile_id)
        """
        response = self.session.delete(
            datafile_urls.datafiles_instance.format(id=datafile_id)
        )

    def get_s3_datafile(self, datafile_id):
        """
        .. deprecated:: 1.8.0

        Fetches an S3 datafile providing easy access to directly
        stream/load files without the need of downloading them.

        If the datafile is not stored in S3, or you don't have access to it
        then an error will be displayed.

        :param str datafile_id: The id of the datafile we want to load

        :returns: a datafile that can read files from S3
        :rtype: dli.client.s3.S3DatafileWrapper

        .. code-block:: python

                datafile = client.get_s3_datafile(datafile_id)
                with datafile.open_file("path/to/file/in/datafile") as f:
                    f.read() # do something with the file

                # or if you want a pandas dataframe created from it you can
                pd.read_csv(datafile.open_file("path/to/file/in/datafile"))

                # you can see all the files in the datafile by calling `files`
                datafile.files  # displays a list of files in this datafile

        """
        warnings.warn(
            '.get_s3_datafile is deprecated. '
            'Please use dataset.download() instead.',
            DeprecationWarning
        )

        datafile = self.get_datafile(datafile_id)
        s3_access = Client.from_dataset(self, datafile.dataset_id)
        return S3DatafileWrapper(datafile._asdict(), s3_access)

    def download_datafile(self, datafile_id, destination, flatten=False):
        """
        Helper function that downloads all files
        registered in a datafile into a given destination.

        This function is only supported for data-lake managed s3 buckets,
        otherwise an error will be displayed.

        Currently supports:
          - s3

        :param str datafile_id: The id of the datafile we want to download files from.

        :param str destination: required. The path on the system, where the
            files should be saved. must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :returns: a list of the paths where the files were downloaded
            to e.g. ['path/git /package/dataset/as_of_date=2019-10-16
            /file.csv.gz'].
            with flatten=True e.g. ['path/file.csv.gz']

        - **Sample**

        .. code-block:: python

                client.download_datafile(datafile_id, destination)
        """
        # get the s3 keys
        # this requires access to be granted
        datafile = self._get_datafile(datafile_id)
        s3_access = Client.from_dataset(self, datafile.datasetId)
        # for each file/folder in the datafile, attempt to download the file
        # rather than failing at the same error, keep to download as much as
        # possible and fail at the end.
        failed = []
        files = [f["path"] for f in datafile.files]
        filepaths = []
        for file in files:
            try:
                download_path = s3_access.download_files_from_s3_path(
                    s3_path=file,
                    destination=destination,
                    flatten=flatten,
                )
                filepaths.extend(download_path)
            except Exception:
                logger.exception(
                    "Failed to download file `%s` from datafile `%s`", file, datafile_id)
                failed.append(file)

        if failed:
            raise DownloadFailed(
                "Some files in this datafile could not be downloaded, "
                "see logs for detailed information. Failed:\n%s"
                % "\n".join(failed)
            )
        return filepaths

    def _process_s3_upload(self, files, s3_location, dataset_id):
        s3_client = Client.from_dataset(self, dataset_id)
        # Ensure trailing slash is included if missing
        s3_location = posixpath.join(s3_location, '')
        return s3_client.upload_files_to_s3(files, s3_location)

    def get_datafile(self, datafile_id):
        """
        Fetches datafile metadata for an existing datafile.

        :param str datafile_id: the unique id of the datafile we want to fetch.

        :returns: The datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.get_datafile(datafile_id)
        """
        datafile = self._get_datafile(datafile_id)

        return siren_to_entity(datafile)

    def _get_datafile(self, datafile_id):
        response = self.session.get(
            datafile_urls.datafiles_instance.format(id=datafile_id)
        )

        return response.to_siren()

    def add_files_to_datafile(self, datafile_id, s3_prefix, files):
        """
        Upload files to existing datafile.

        :param str datafile_id: The id of the datafile to be updated.
        :param str s3_prefix: Location for the files in the destination s3 bucket. Cannot end with slash ("/").
        :param list[str] files: List of files to be added to the datafile.

        :returns: The updated datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile_updated_with_additional_files = client.add_files_to_datafile(
                                                              datafile_id=datafile_id,
                                                              s3_prefix="quotes/20180518",
                                                              files=["./data/AAPL.csv", "./data/MSFT.csv"],
                                                        )
        """
        datafile = self.get_datafile(datafile_id)
        dataset = self.get_dataset(id=datafile.dataset_id)

        if dataset.location.type != "S3":
            raise Exception("Can not upload files to non-S3 datasets.")

        s3_location = "{}/{}".format(dataset.location.bucket, s3_prefix)
        uploaded_files = self._process_s3_upload(
            files,
            s3_location,
            datafile.dataset_id
        )

        if datafile.files:
            uploaded_files.extend(datafile.files)

        return self.edit_datafile_metadata(datafile_id, files=uploaded_files)

    @staticmethod
    def _ensure_iso_date(date):
        if date:
            return datetime.datetime.strptime(date, "%Y-%m-%d").strftime("%Y-%m-%d")
        return date



from contextlib import redirect_stdout

import pytest
import copy
import os
import io
import pytest
import tempfile
import pyarrow
import pandas
import numpy
import random
import json

from dli.client.exceptions import DataframeStreamingException
from unittest.mock import MagicMock, ANY, patch


@pytest.fixture
def instance_file(client):
    client._environment.consumption = 'consumption-test.local'

    instance = client.Instance(
        datafile_id='123'
    )

    client._session.get().json.return_value = {
        'data': [{
            'attributes': {
                'path': 'prefix/test.001.csv',
                'metadata': {}
            }
        }]
    }

    files = instance.files()
    file_ = files[0]

    client.session.reset_mock()

    client._session.get().raw.read.side_effect = [
        b'hello world',
        b''
    ]

    yield file_


class TestDataset:

    def test_dataset_instances_is_iterable(self, client):
        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        })

        datafiles = dataset.instances.all()
        assert hasattr(datafiles, '__iter__')

    def test_dataset_download_refers_to_latest_instance(self, client):
        path = './destination_path'
        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        })

        dataset.instances.latest = MagicMock(
            download=MagicMock(return_value='file')
        )

        result = dataset.instances.latest.download(path)

        dataset.instances.latest.download.assert_called_once_with(path)
        assert result == 'file'

    def test_dataset_instances_make_requests_on_iteration(self, client):
        client._session.get().json.side_effect = [
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 1}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 2}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 3}
                }]
            },
        ]
        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        })

        instances = list(dataset.instances.all())
        assert len(instances) == 3

    def test_dataset_instances_make_multiple_iterations(self, client,
                                                        monkeypatch):
        client._session.get().json.side_effect = [
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 1}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 2}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 3}
                }]
            },
        ]
        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        })

        def mock_iter(self):
            if not hasattr(mock_iter, 'call_count'):
                mock_iter.call_count = 0
                mock_iter.cached = False

            if not mock_iter.cached:
                mock_iter.call_count += 1
                mock_iter.cached = True

            yield from [1, 2, 3]

        instances = list(dataset.instances.all())
        assert len(instances) == 3

        instances_from_cache = list(dataset.instances.all())
        assert(len(dataset.instances._paginator.cache) == 3)
        assert(len(instances_from_cache) == 3)

    def test_dataset_filter_prefix(self, client, monkeypatch, instance_file):
        boto_mock = MagicMock()
        monkeypatch.setattr(
            'dli.models.dataset_model.create_refreshing_session',
            boto_mock
        )

        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        with tempfile.TemporaryDirectory() as target_dest:
            dataset.download(target_dest)

        assert (
            boto_mock().resource().Bucket()
            .objects.filter.call_args[1]['Prefix']
        ) == 'abc/'

        with tempfile.TemporaryDirectory() as target_dest:
            dataset.download(target_dest, filter_path='/123')

        assert (
            boto_mock().resource().Bucket()
            .objects.filter.call_args[1]['Prefix']
        ) == 'abc/123'

    def test_dataset_is_wrapped_with_analytics_logger(self, client,
                                                      monkeypatch,
                                                      instance_file):
        boto_mock = MagicMock()
        monkeypatch.setattr('dli.aws.Session', boto_mock)

        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        with tempfile.TemporaryDirectory() as target_dest:
            dataset.download(target_dest)

        client._analytics_handler.create_event.call_count = 2
        client._analytics_handler.create_event.assert_any_call(
            # Not testing the extract user metadata here
            ANY, ANY,
            'DatasetModel',
            'download',
            {
                'self': dataset,
                'destination_path': target_dest,
                'dataset_id': '1234'
            },
            result_status_code=200
        )
    def test_dataset_info(self, test_client, dataset_request, dictionary_request, fields_request):
        test_client.session.reset_mock()
        test_client._session.get.return_value.json.side_effect = [
            dictionary_request, fields_request
        ]

        dataset = test_client.Dataset._from_v2_response(dataset_request)

        def redirect(fun):
            with io.StringIO() as buf, redirect_stdout(buf):
                fun()
                output = buf.getvalue()
                return output


        assert(
            redirect(dataset.info)
                .replace(" ", "").replace("\n", "")
               ==
            """
                name    type
                ------  -----------------
                col1    String (Nullable)
                col2    String (Nullable)
                col3    String
            """
                .replace(" ", "").replace("\n", "")
        )


class TestFileDownload:

    def test_instance_file_download_simple(self, client, instance_file):
        old_dir = os.path.abspath(os.curdir)

        with tempfile.TemporaryDirectory() as target_dest:
            os.chdir(os.path.abspath(target_dest))
            instance_file.download()
            with open(os.path.join(target_dest, 'prefix/test.001.csv')) as f:
                assert f.read() == 'hello world'

        os.chdir(old_dir)

    def test_instance_file_download_with_path(self, client, instance_file):
        with tempfile.TemporaryDirectory() as target_dest:
            instance_file.download(target_dest)
            with open(os.path.join(target_dest, 'prefix/test.001.csv')) as f:
                assert f.read() == 'hello world'

    def test_instance_file_download_non_existent_path(
        self, client, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a/b/c/')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

    def test_instance_file_download_non_existent_file(
        self, client, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

    def test_instance_file_overwrites(
        self, client, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

        client.session.reset_mock()
        client._session.get().raw.read.side_effect = [
            b'brave new world',
            b''
        ]

        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'brave new world'


class TestInstanceDownload:

    @pytest.fixture
    def instance(self, client):
        client._environment.consumption = 'consumption-test.local'

        instance = client.Instance(
            datafile_id='123'
        )

        client._session.get().json.return_value = {
            'data': [{
                'attributes': {
                    'path': f's3://bucket/prefix/test.00{i}.csv',
                    'metadata': {}
                }
            } for i in range(3)
            ]
        }

        yield instance

    def test_instance_download_all_should_refer_to_download_function(self, instance):
        path = './target'
        instance.download = MagicMock(return_value='file')
        result = instance.download_all(path, flatten=False)

        instance.download.assert_called_once_with(path, False)
        assert result == 'file'

    def test_instance_download_flatten(self, client, monkeypatch, instance):
        client._session.get().raw = io.BytesIO(b'abc')

        with tempfile.TemporaryDirectory() as target_dest:
            to = os.path.join(target_dest, 'a')
            result = instance.download_all(to, flatten=True)
            directory_of_flattened = result[0].split('/test')[0]
            list_dir = sorted(os.listdir(directory_of_flattened))

            assert (
                list_dir ==
                ['test.000.csv', 'test.001.csv', 'test.002.csv']
            )


class TestInstanceDataframe:

    @pytest.fixture
    def test_dataframe(self) -> pandas.DataFrame:
        # Rows:
        # 100 -> 205.5 kB (205,548 bytes) on disk in 0.44 seconds.
        # 10000 -> 9.2 MB (9,234,131 bytes) on disk in 0.67 seconds.

        max_rows = 10000
        # Max columns should match a worst case. For CDS Single Name Pricing
        # Sensitivities and Liquidity there are 164 columns.
        max_columns = 168

        num_column_types = 6

        # Seed the random number generator so that we can reproduce the output.
        numpy.random.seed(0)

        def bools_df():
            return pandas.DataFrame(
                numpy.random.choice(
                    [False, True],
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_bool_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def ints_df():
            return pandas.DataFrame(
                numpy.random.randint(
                    low=0,
                    high=1000000,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_int_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def floats_df():
            return pandas.DataFrame(
                numpy.random.rand(
                    max_rows,
                    int(max_columns / num_column_types)),
                columns=[str(f'column_float_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def string_df():
            import pandas.util.testing
            return pandas.DataFrame(
                numpy.random.choice(
                    pandas.util.testing.RANDS_CHARS,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_str_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def dates_df():
            dates = pandas.date_range(start='1970-01-01', end='2020-01-01')

            return pandas.DataFrame(
                numpy.random.choice(
                    dates,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_date_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def datetimes_df():
            datetimes = pandas.date_range(start='1970-01-01', end='2020-01-01',
                                          freq='H')

            return pandas.DataFrame(
                numpy.random.choice(
                    datetimes,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_datetime_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        df = pandas.concat(
            [bools_df(), ints_df(), floats_df(), string_df(), dates_df(),
             datetimes_df()],
            # Concat columns axis
            axis=1,
            copy=False
        )

        # Randomise the column order so the compression cannot optimise
        # easily based on types as this will mix float columns among the int
        # columns.
        random_column_order = df.columns.tolist()
        random.shuffle(random_column_order)
        df = df[random_column_order]

        assert df.shape[0] == max_rows
        assert df.shape[1] == max_columns

        return df

    @pytest.fixture
    def arrow_ipc_stream(self, test_dataframe):
        sink = pyarrow.BufferOutputStream()
        batch = pyarrow.RecordBatch.from_pandas(test_dataframe)
        writer = pyarrow.RecordBatchStreamWriter(sink, batch.schema)
        writer.write_batch(batch)
        # It's important the IPC stream is closed
        writer.close()
        return io.BytesIO(sink.getvalue().to_pybytes())

    @pytest.fixture
    def dataset(self, client):
        return client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                    'has_access': True,
                }
            }
        })

    def test_dataset_dataframe(self, benchmark, client, arrow_ipc_stream,
                               test_dataframe, dataset):

        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'
        dataset._client.session.get().raw = arrow_ipc_stream
        dataset._client.session.get().status_code = 200

        def _bench():
            # reset our fake socket
            arrow_ipc_stream.seek(0)
            return dataset.dataframe()

        # The benchmark code will run several times to get an average runtime.
        df = benchmark(_bench)

        # Check that we can read from the file that was written.
        pandas.testing.assert_frame_equal(test_dataframe, df)

    @pytest.mark.parametrize('value', ['1', 0])
    def test_dataframe_nrows_validation(self, client, arrow_ipc_stream,
                                        test_dataframe, value, dataset):
        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'
        dataset._client.session.get().raw = arrow_ipc_stream
        dataset._client.session.get().status_code = 200

        # Expected to pass.
        df = dataset.dataframe(value)
        # Check that we can read from the file that was written.
        pandas.testing.assert_frame_equal(test_dataframe, df)

    @pytest.fixture
    def partitions(self):
        return {
            'data':{
                'attributes': {
                    'partitions': {
                        'a': ["1", "2", "3"],
                        'b': ["4", "5", "6"]
                    }
                }
            }
        }

    def test_dataset_partitions(self, client, partitions):
        dataset = client.Dataset._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        })

        dataset._client = client
        dataset._client._session.get.return_value.json.side_effect = [partitions, {
            'data': {
                'id': '1234',
                'attributes': {
                    'location': {},
                }
            }
        },  partitions]

        msg = dataset.partitions()
        assert(msg == partitions["data"]["attributes"]["partitions"])

        msg = client.get_dataset_partitions(id=dataset.id)
        assert(msg == partitions["data"]["attributes"]["partitions"])


    def test_dataframe_handle_errors(self, client, arrow_ipc_stream,
                                     test_dataframe, dataset):
        error_msg = json.dumps({
            'status': 400,
        }).encode('utf')

        arrow_ipc_stream.seek(0, 2)
        arrow_ipc_stream.write(error_msg)
        arrow_ipc_stream.seek(0)

        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'
        dataset._client.session.get().raw = arrow_ipc_stream
        dataset._client.session.get().status_code = 200

        with pytest.raises(DataframeStreamingException):
            dataset.dataframe(1)


@pytest.mark.xfail(reason='The packages() and datasets() endpoints are disabled')
class TestDatasetModule:

    def test_dataset_contents(capsys, test_client, package_request,
                             package_request_v2, dataset_request_index,
                             instance_request):
        test_client._session.get.return_value.json.side_effect = [
            package_request, package_request_v2,
            dataset_request_index, instance_request,
        ]

        pp = test_client.packages(search_term='', only_mine=False)
        ds = pp['Test Package'].datasets()
        d = ds['Test DataSet']
        d.contents()
        captured = capsys.readouterr()
        assert captured.out == '\nINSTANCE 1\n'

    def test_retrieve_datasets(test_client, dataset_request_v1,
                               dataset_request):
        test_client._session.get.return_value.json.side_effect = [
            dataset_request_v1, dataset_request
        ]

        # calls in the same way as package
        pp = test_client.datasets(search_term='', only_mine=False)
        assert len(pp) == 1

    def test_datasets_called_twice(test_client, package_request,
                                   package_request_v2, dataset_request_index):
        test_client._session.get.return_value.json.side_effect = [
            package_request, package_request_v2, dataset_request_index
        ]
        pp = test_client.packages(search_term='', only_mine=False)

        # This should work twice
        assert pp['Test Package'].datasets()
        assert pp['Test Package'].datasets()

    def test_retrieve_instance_given_package_dataset(test_client,
                                                     package_request,
                                                     package_request_v2,
                                                     dataset_request_index,
                                                     instance_request
                                                     ):
        test_client._session.get.return_value.json.side_effect = [
            package_request, package_request_v2,
            dataset_request_index, instance_request,
        ]

        pp = test_client.packages(search_term='', only_mine=False)
        ds = pp['Test Package'].datasets()
        d = ds['Test DataSet']
        instances = list(d.instances.all())
        assert (len(instances) == 1)
        assert (instances[0].datafile_id == 1)



import pytest

from dli.client.dli_client import DliClient
from datetime import datetime
from unittest.mock import MagicMock


class TestClientMock(DliClient):

    def __init__(self, args, kwargs):
        self._session = MagicMock()
        self._session.auth_key = 'abc'
        self._session.token_expires_on = datetime(2100, 1, 1)
        self._environment = MagicMock()
        self._environment.consumption = ''
        self._environment.s3_proxy = 's3proxy.fake'
        self._analytics_handler = MagicMock()
        self.packages = self._packages()
        self.datasets = self._datasets()

    @property
    def session(self):
        return self._session


@pytest.fixture
def test_client():
    yield TestClientMock('abc', '123')


@pytest.fixture
def client():
    yield TestClientMock('abc', '123')


@pytest.fixture
def package_request():
    return {
            'class': ['Access Requests', 'collection'],
            'properties': {'page': 1, 'pages_count': 1},
            'entities': [{
                'properties': {
                    'packageId': '1234',
                    'name': 'Test Package',
                    'data': {
                        'name': 'Test Package',
                        'topic': 'Test Topic',
                        'keywords': ['test'],
                        'hasAccess': True,
                        'description': 'DESC',
                        'documentation': 'href',
                        'dataSensitivity': 'Private',
                    },
                    'total_records': 1,
                }
            }]
           }

@pytest.fixture
def package_request_v2():
    return {
        'data': {
            'id': '1234',
            'attributes': {
                'name': 'Test Package',
                'topic': 'Test Topic',
                'keywords': ['test'],
                'hasAccess': True,
                'description': 'DESC',
                'documentation': 'href',
                'dataSensitivity': 'Private',
            }
        }
    }

@pytest.fixture
def package_request_v2():
    return {
        'data': {
            'id': '1234',
            'attributes': {
                'name': 'Test Package',
                'topic': 'Test Topic',
                'keywords': ['test'],
                'hasAccess': True,
                'description': 'DESC',
                'documentation': 'href',
                'dataSensitivity': 'Private',
            }
        }
    }


@pytest.fixture
def dataset_request_index():
    return {
            'data':
                [{'type': 'dataset',
                 'attributes': {
                     'package_id': '59e6a698-975d-11e9-84cc-7e5ef76d533f',
                     'updated_by': 'update@ihsmarkit.com',
                     'keywords': [],
                     'location': {},
                     'has_access': False,
                     'first_datafile_at': '2019-05-06',
                     'short_code': 'TestDataset',
                     'absolute_taxonomy': 'Data-Test',
                     'publishing_frequency': 'Monthly',
                     'name': 'Test DataSet',
                     'last_datafile_at': '2020-01-01',
                     'data_format': 'PARQUET',
                     'created_at': '2019-06-25',
                     'description': 'description',
                     'organisation_name': 'IHS Markit',
                     'created_by': 'creator@ihsmarkit.com',
                     'taxonomy': []
                 },
                 'id': '5b01376e-975d-11e9-8832-7e5ef76d533f'
                 }],
            'meta': {'total_count': 1, 'page': 1,
                     'total_pages': 1}
          }


@pytest.fixture
def dataset_request():
    return {
            'data': {
                'type': 'dataset',
                'attributes': {
                     'package_id': '59e6a698-975d-11e9-84cc-7e5ef76d533f',
                     'updated_by': 'update@ihsmarkit.com',
                     'keywords': [],
                     'location': {},
                     'has_access': False,
                     'first_datafile_at': '2019-05-06',
                     'short_code': 'TestDataset',
                     'absolute_taxonomy': 'Data-Test',
                     'publishing_frequency': 'Monthly',
                     'name': 'Test DataSet',
                     'last_datafile_at': '2020-01-01',
                     'data_format': 'PARQUET',
                     'created_at': '2019-06-25',
                     'description': 'description',
                     'organisation_name': 'IHS Markit',
                     'created_by': 'creator@ihsmarkit.com',
                     'taxonomy': []
                 },
             'id': '5b01376e-975d-11e9-8832-7e5ef76d533f'
        }
      }


@pytest.fixture
def dataset_request_v1():
    return {
        'data': [{
            'class': None,
            'title': None,
            'properties': {
                'name': None,
                'keywords': None,
                'loadType': None,
                'location': None,
                'taxonomy': None,
                'createdAt': None,
                'createdBy': None,
                'datasetId': None,
                'packageId': None,
                'updatedAt': None,
                'updatedBy': None,
                'dataFormat': None,
                'contentType': None,
                'description': None,
                'documentation': None,
                'lastDatafileAt': None,
                'firstDatafileAt': None,
                'namingConvention': None,
                'publishingFrequency': None,
                'hasDatafileMonitoring': None,
                'shortCode': None,
                'managerId': None,
                'techDataOpsId': None,
                'accessManagerId': None,
                'tenantId': None,
                'tenantName': None,
                'isInternalWithinOrganisation': None,
                'type': None,
                'tags': None,
                'index_join_field': None,
                'packageName': None,
                'hasAccess': None
                },
            'package_id': '59e6a698-975d-11e9-84cc-7e5ef76d533f',
            'updated_by': 'update@ihsmarkit.com',
            'keywords': [],
            'location': None,
            'links': None,
            'actions': None,
            'rel': None
        }],
        'meta': {
            'total_count': 1, 'page': 1,
            'total_pages': 1}
                   }

@pytest.fixture
def instance_request():
    return {
        'properties': {'pages_count': 1},
        'entities': [{
            'properties': {'datafileId': 1}
        }]
    }

@pytest.fixture
def dictionary_request():
    return {
        'data': {
        'type': 'dictionary',
        'attributes': {
          'dataset_id': '7eb6eb49-5fae-44bc-900a-c241e0a24b00',
          'description': None,
          'version': '1',
          'partitions': [],
          'valid_as_of': '2019-09-26'
        },
        'id': '2d113c3a-be1a-4c70-b87a-8e569dee8dd0'}
    }

@pytest.fixture
def fields_request():
    return {
        'data':
        {
            'type': 'dictionary_fields',
            'attributes': {
                  'fields': [
                    {'nullable': True,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col1'},
                    {'nullable': True,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col2'},
                    {'nullable': False,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col3'},
                  ]
            },
            'id': '2d113c3a-be1a-4c70-b87a-8e569dee8dd0'
        },
        'meta': {'total_count': 3}
    }


import logging
import sys
# from logging.handlers import RotatingFileHandler
# from pathlib import Path

from pythonjsonlogger import jsonlogger  # noqa: I900


def _setup_logging(where_level):
    logger = logging.getLogger(__name__)

    json_format = jsonlogger.JsonFormatter(
        '%(asctime)s - %(levelname)s - %(name)s - %(message)s',
        json_indent=2
    )

    handler = None
    chosen_level = None
    switcher = {
        "error": logging.ERROR,
        "warn": logging.WARNING,
        "debug": logging.DEBUG,
        "info": logging.INFO
    }

    if where_level:
        # If debug - create a log file or stdout
        to, _, level = str(where_level).partition(":")

        if to == "stdout":
            handler = logging.StreamHandler(stream=sys.stdout)

        chosen_level = switcher.get(level.lower(), logging.DEBUG)

    # if not handler:
    #     log_folder = "logs/"
    #     from pathlib import Path
    #     Path(log_folder).mkdir(parents=True, exist_ok=True)
    #     from logging.handlers import RotatingFileHandler
    #     handler = RotatingFileHandler(
    #         f'{log_folder}sdk.log', mode='w', backupCount=3)
    #     chosen_level = logging.DEBUG

    # handler.setFormatter(json_format)
    # handler.setLevel(chosen_level)
    # logger.addHandler(handler)
    logger.setLevel(chosen_level)
    return logger



import copy
import os
import json

from requests_toolbelt.multipart.encoder import MultipartEncoder

TEN_MEGABYTES = 10485760

class DatasetBuilder:
    """
    Dataset is grouping of related datafiles. It provides user with metadata required to consume and use the data.

    This builder object sets sensible defaults and exposes
    helper methods on how to configure its storage options..

    :param str package_id: Package ID to which the dataset belongs to.
    :param str name: A descriptive name of the dataset. It should be unique within the package the dataset
                    belongs to.
    :param str description: A short description of the dataset.
    :param str content_type: A way for the data steward to classify the type of data in the dataset
                        (e.g. pricing).
    :param str data_format: The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `TSV`,`XML`, `Other`.
    :param str publishing_frequency: The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
    :param list[str] taxonomy: A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
    :param str short_code: Optional. A short code for the dataset, should be an alphanumerical camel-cased string.
                        If not provided, a short_code will be generated based on the name.
    :param list[str] keywords: Optional. User-defined list of separated keywords that can be used to find this dataset
                            through the search interface.
    :param str naming_convention: Optional. Key for how to read the dataset name.
    :param str documentation: Optional. Documentation about this dataset in markdown format.
    :param str load_type: Optional. Whether each datafile in this dataset should be considered as a full version of a
                        dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`
    :param str has_datafile_monitoring: Optional. Notify when latest instance is missing

    # XXX
    :param str data_preview_type: Optional enum of NONE STATIC LIVE. 'NONE' means there is no datapreview. STATIC means
                        to use the file provided. LIVE means to use the latest datafile.
    :param file sample_data: Optional file. File like object. Used in conjuntion with data_preview_type.
    """

    @staticmethod
    def _create_dliv1_paylod(
        package_id=None,
        name=None,
        description=None,
        content_type=None,
        data_format=None,
        publishing_frequency=None,
        taxonomy=None,
        keywords=None,
        naming_convention=None,
        documentation=None,
        load_type=None,
        **kwargs
    ):
        return {
            "packageId": package_id,
            "name": name,
            "description": description,
            "keywords": keywords,
            "contentType": content_type,
            "location": None,
            "dataFormat": data_format,
            "publishingFrequency": publishing_frequency,
            "namingConvention": naming_convention,
            "documentation": documentation,
            "taxonomy": taxonomy,
            "loadType": load_type
        }

    def __init__(self, location_builder=None, **kwargs):
        self._data = dict(kwargs)

        self._data['sample_data'] = self._data.get('sample_data')
        if self._data.get('sample_data'):
            self._data['data_preview_type'] = (
                    self._data.get('data_preview_type') or 'STATIC'
            )
            self._validate_preview_type_static(
                self._data.get('data_preview_type')
            )

        # this is just for DLI v1 TODO add deprecated to it.
        self.payload = self._create_dliv1_paylod(**self._data)
        self._location_builder = location_builder

    def to_json_api(self):
        attributes = dict(**self._data)
        if self._location_builder is not None:
            attributes.update(**self._location_builder._dliv2())

        # This is a file
        attributes.pop('sample_data', None)
        return {
            'type': 'dataset',
            'data': {'attributes': attributes}
        }

    def to_multipart_body(self):
        data = json.dumps(self.to_json_api())

        fields={
            'data': ('data', data, 'application/json'),
        }

        if self._data['sample_data']:
            fields['sample_data'] = (
                os.path.basename(self._data['sample_data'].name),
                self._data['sample_data'],
                # Mimetyppe is ignored
                'text/plain'
            )

        encoder = MultipartEncoder(fields=fields)

        # NOTE this really should be fixed on the catalogue as well.
        # the way to handle this and deliver an error message more
        # graceful than an arbitary socket closure is for the catalogue
        # to reject the requests with a 'Content-Length' header that's
        # too large.
        if encoder.len > (TEN_MEGABYTES - len(data)):
            raise ValueError(
                f'The sample_data file passed to the'
                ' DatasetBuilder is too large. Please '
                'keep files under 10 megabytes.'
            )

        return encoder

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A valid path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").
        :returns: itself
        :rtype: dli.client.builders.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Pricing",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                                data_preview_type='STATIC',
                                sample_data=open('path/to/my/file.csv')
                        )
                builder = builder.with_external_s3_storage(
                    bucket_name="external-s3-bucket-name",
                    aws_account_id=123456789,
                    prefix="/economic-data-package/my-test-dataset"
                )
                dataset = client.register_dataset(builder)
        """
        self._location_builder = DatasetLocationBuilder().with_external_s3_storage(
            bucket_name=bucket_name,
            aws_account_id=aws_account_id,
            prefix=prefix
        )
        self.payload.update(self._location_builder.build())
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :returns: itself
        :rtype: dli.client.builders.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Pricing",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                        )
                builder = builder.with_external_storage("external-storage-location")
                dataset = client.register_dataset(builder)
        """
        self._location_builder = DatasetLocationBuilder().with_external_storage(location)
        self.payload.update(self._location_builder.build())
        return self

    def build(self):
        return copy.copy(self.payload)

    @staticmethod
    def _validate_preview_type_static(preview_type):
        if preview_type != 'STATIC':
            raise ValueError(
                'Field: data_preview_type must be STATIC when '
                'sample_data is provided'
            )


class DatasetLocationBuilder:
    """
        A simple builder to specify dataset location.
    """

    def __init__(self):
        self.payload = {}

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A vaild path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").

        :returns: itself
        :rtype: dli.client.builders.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_s3_storage(
                        bucket_name="external-s3-bucket-name",
                        aws_account_id=123456789,
                        prefix="/economic-data-package/my-test-dataset"
                    )
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "S3",
            "owner": {
                "awsAccountId": str(aws_account_id)
            },
            "bucket": bucket_name,
            "prefix": prefix
        }
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :returns: itself
        :rtype: dli.client.builders.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "Other",
            "source": location
        }
        return self

    def _dliv2(self):
        """
        The old API used a polymorphic location type.
        The new API uses keys. While more simple we
        have to handle both cases.
        """
        data = {}

        if self.payload['location']['type'] == 'S3':
            data['location'] = {
                's3': copy.copy(self.payload['location'])
            }

            if 'owner' in data['location']['s3']:
                data['location']['s3']['owner']['aws_account_id'] = (
                    # Is camel cased in old API but not new api...
                    data['location']['s3']['owner'].pop('awsAccountId')
                )
        else:
            data['location'] = {
                'other': copy.copy(self.payload['location'])
            }

        return data

    def build(self):
        return copy.copy(self.payload)



import base64
import hashlib
import json
import logging
import os
import re
import sys
import socket
import threading
import uuid
from urllib.parse import urlsplit, parse_qs, urlunparse, urljoin
import furl
import requests
from flask import Flask, Blueprint, redirect, request, session
from requests.exceptions import ConnectTimeout

from dli.client.components.urls import sam_urls, identity_urls


def can_launch(port):
    current_port = port
    # if the user opens ipython in multiple windows, the first will
    # be using the 8080 socket, and then the second will not launch a
    # new server. Because we get the value using a wait/notify against
    # Listener.values global, we must use an open port, else the wrong
    # server instance's Listener.values global will be set.
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # already_open = (result == 0)
        # reason = socket.errno.errorcode.get(result, "SUCCESS")
        try:
            s.bind(("localhost", current_port))
            result = True, "Bound"
        except Exception as e:
            result = False, str(e)
        finally:
            s.close()

        return result


class _Listener:

    DEFAULT_PORT = 8080
    LOCALHOST = "http://localhost"
    VALUES = {}
    _routes = None
    # _lock = threading.RLock()
    # condition = threading.Condition(_lock)

    #disable the flask startup messaging
    cli = sys.modules['flask.cli']
    cli.show_server_banner = lambda *x: None
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)


    @staticmethod
    def setup_verifiers(code):
        code_verifier = base64.urlsafe_b64encode(code).decode(
            'utf-8')
        code_verifier = re.sub('[^a-zA-Z0-9]+', '', code_verifier)
        code_challenge = hashlib.sha256(code_verifier.encode('utf-8')).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode(
            'utf-8')
        code_challenge = code_challenge.replace('=', '')
        return code_challenge, code_verifier

    @staticmethod
    def register_routes(app, port):
        localhost = f"{_Listener.LOCALHOST}:{port}"
        _Listener._routes = Blueprint("auth", __name__, )

        @_Listener._routes.route('/login', methods=['GET'])
        def login():
            postbox = request.args.get("postbox")
            if not postbox:
                return """
                  <img style="max-width: 300px; height: auto; "
                    src="https://cdn.ihsmarkit.com/www2/a/p/media/images/ihsmarkit.svg" 
                    alt="IHS Logo"
                  >
                  </hr>
                  <h2 style="font-family: Arial, Helvetica, sans-serif;">
                  Forbidden. 
                  </h2>
                """, 403

            try:
                # setup the session
                rnd = os.urandom(40)
                session["catalogue"] = request.args.get("catalogue")
                session["sam_client"] = request.args.get("sam_client")
                session["sam"] = request.args.get("sam")
                session["code_challenge"], session["code_verifier"] = \
                    _Listener.setup_verifiers(rnd)
                session["postbox"] = str(postbox)

                response = requests.head(
                    urljoin(session["catalogue"], "login"),
                    timeout=10
                )
                bits = urlsplit(response.headers["Location"])
                params = parse_qs(bits.query)
                params["redirect_uri"] = [localhost]
                path = urlunparse((
                    bits.scheme, bits.netloc, bits.path, '', '', ''
                ))

                target_builder = furl.furl(path)

                target_builder.args = {
                    "state": postbox,
                    "client_id": session["sam_client"],
                    "response_type": "code",
                    "redirect_uri": localhost,
                    "scope": "openid profile saml_attributes email",
                    "code_challenge": session["code_challenge"],
                    "code_challenge_method": "S256"
                }
                redi = redirect(target_builder.url)
                return redi
            except ConnectTimeout:
                return "Could not connect - check internet connection", 400
            except Exception as e:
                return str(e), 500

        @_Listener._routes.route('/', methods=['GET', 'POST'])
        def auth_callback():

            postbox = session["postbox"]
            logging.info(f"Current postbox {postbox}")
            state = request.args.get('state', '')
            if state != postbox:
                # we have to set this, else the shell will just hang waiting
                # for a value - so they must check it and quit etc.
                _Listener.VALUES[postbox] = "invalid"
                return """
                  <img style="max-width: 300px; height: auto; "
                    src="https://cdn.ihsmarkit.com/www2/a/p/media/images/ihsmarkit.svg"
                    alt="IHS Logo"
                  >
                  </hr>
                  <h2 style="font-family: Arial, Helvetica, sans-serif;">
                  You have another SDK session open.
                  </h2>
                """, 403

            code = request.args.get('code')
            client_id = request.args.get('client_id')

            # exchange this code for a token
            tokens = requests.post(
                urljoin(session["sam"], sam_urls.sam_token),
                data={
                    "grant_type": "authorization_code",
                    "client_id": client_id,
                    "redirect_uri": localhost,
                    "code": code,
                    "code_verifier": session["code_verifier"]
                },
                allow_redirects=False
            )

            if tokens.status_code != 200:
                return tokens.text, tokens.status_code


            token = tokens.json()["id_token"]
            # ok now we wanna exchange this token with catalogue
            catalogue_response = requests.post(
                urljoin(session["catalogue"],
                        identity_urls.identity_token),
                data={
                    "client_id": client_id,
                    "subject_token": token,
                    "origin": "SDK"

                },
                allow_redirects=False
            )


            if catalogue_response.status_code != 200:
                return catalogue_response.text, catalogue_response.status_code

            jwt = catalogue_response.json()["access_token"]
            # with _Listener.condition:
            # logging.info(f"We are setting the postbox {postbox}")
            _Listener.VALUES[postbox] = jwt
            # _Listener.condition.notify_all()

            logging.info("Acquired Catalogue JWT")
            return """
              <img style="max-width: 300px; height: auto; "
                src="https://cdn.ihsmarkit.com/www2/a/p/media/images/ihsmarkit.svg" 
                alt="IHS Logo"
              >
              </hr>
              <h2 style="font-family: Arial, Helvetica, sans-serif;">
              You're now logged in. Please close this window.
              </h2>
            """

        @_Listener._routes.route('/shutdown', methods=['GET', 'POST'])
        def shutdown():
            try:
                pass
            finally:
                return 'Server shutting down...'

        @_Listener._routes.route('/get/<query_postbox>')
        def get(query_postbox):
            if _Listener.VALUES.get(query_postbox, False):
                return json.dumps({
                    "jwt": _Listener.VALUES.get(query_postbox, "")
                }), 200

            return json.dumps({"jwt": "None"}), 404


        app.register_blueprint(_Listener._routes)

    @staticmethod
    def run(port=DEFAULT_PORT, debug=False):
        postbox = base64.urlsafe_b64encode(uuid.uuid4().bytes)\
            .decode('utf-8').replace('=', '')

        app = Flask(__name__)
        app.logger.disabled = True
        app.secret_key = os.urandom(20)
        if debug:
            _Listener.app = app

        _Listener.register_routes(app, port)

        # check whether we need to launch the listener server
        if not debug:
            result, reason = can_launch(_Listener.DEFAULT_PORT)
            if result:
                try:
                    t = threading.Thread(
                        target=lambda: app.run(port=port)
                    )
                    t.daemon = True
                    t.start()
                except OSError:
                    print("Already running")

        return str(postbox)



import os
import warnings
import requests

from dli.client import _setup_logging
from dli.client.dli_client import DliClient
from dli import __version__

DEPRECATED_AGE = 1

def version_check(package_name):
    url = f"https://pypi.python.org/pypi/{package_name}/json"

    try:
        data = requests.get(url).json()
        versions = list(x for x in data["releases"].keys() if 'b' not in x)
        versions = sorted(versions)
        versions.reverse()

        offset = versions.index(__version__)

        if offset > DEPRECATED_AGE:
            warnings.warn("You are using an old version of the SDK, please upgrade"
                          "using `pip install [dli] --upgrade` "
                          "before the SDK no longer functions as expected",
                          PendingDeprecationWarning)
    except:
        # Version check has been see to fail by Gilberto with an SSLError
        # when doing the handshake with pypi.
        pass


def start_session(
    api_key=None,
    root_url="https://catalogue.datalake.ihsmarkit.com/__api",
    host=None,
    debug=False,
    strict=True,
    use_keyring=True
):
    """
    Entry point for the Data Lake SDK, returns a client instance that
    can be used to consume or register datasets.

    Example for starting a session:

        from dli.client import session
        dl = session.start_session(api_key)

    :param str api_key: Your API key, can be retrieved from your dashboard in
                        the Catalogue UI.
    :param str root_url: Optional. The environment you want to point to. By default it
                        points to Production.
    :param str host: Optional. Advanced usage, meant to force a hostname when DNS resolution
                     is not reacheable from the user's network.
                     This is meant to be used in conjunction with an
                     IP address as the root url.
                     Example: catalogue.datalake.ihsmarkit.com

    :param bool debug: Optional. Log SDK operations to a file in the current working
                       directory with the format "sdk-{end of api key}-{timestamp}.log"

    :param bool strict: Optional. When True, all exception messages and stack
                        trace are printed. When False, a shorter message is
                        printed and `None` should be returned.
    :param bool use_keyring: Optional. When True, cache the JWT in the system keyring
                             and retrieve that JWT from the system keyring if set.
                             Otherwise disable the keyring completely.

    :returns: Data Lake interface client
    :rtype: dli.client.dli_client.DliClient

    """
    logger = _setup_logging('stdout:')
    version_check('dli')

    user = os.environ.get("DLI_ACCESS_KEY_ID")
    pasw = os.environ.get("DLI_SECRET_ACCESS_KEY")
    api_key = api_key or os.environ.get("DLI_API_KEY")

    logger.debug("Checking auth flow")

    if (user is not None and pasw is not None) and api_key is None:
        logger.debug("Using SAM client credentials authentication flow")
        return get_client()(root_url, host, debug=debug, strict=strict,
                            access_id=user, secret_key=pasw, use_keyring=use_keyring)
    elif (user is None or pasw is None) and api_key is None:
        logger.debug("Using SAM PKCE web authentication flow")
        return get_client()(root_url, host, debug=debug, strict=strict, use_keyring=use_keyring)
    elif api_key is not None:
        logger.debug("Using API Key authentication flow")
        if api_key is not None:
            return get_client()(root_url, host, debug=debug, strict=strict,
                                secret_key=api_key, use_keyring=use_keyring)

def get_client():
    return DliClient



import os
import warnings

warnings.filterwarnings(
    'always', module="dli"
)

warnings.filterwarnings(
    'ignore', module="dli", category=ResourceWarning
)

__version__ = '1.8.4b6'
__product__ = "ihsm-datalake"


try:
    import simplejson as _  # noqa
    warnings.warn(
        'Incompatible Package `simplejson`.\n\n'
        '\t`simplejson` is a backport of the built in json library in Python. '
        'It contains subtle differences, and is not intended for use beyond '
        'Python 2.6. Please uninstall `simplejson` by running:\n\n'
        '\t\tpip uninstall simplejson\n\n'
        '\tOr run the DLI from a virtual environment as it is known to cause '
        'issues within the DLI.\n',
        ImportWarning
    )
except ImportError:
    pass


def connect(api_key=None,
            root_url="https://catalogue.datalake.ihsmarkit.com/__api",
            host=None,
            debug=None,
            strict=None,
            use_keyring=True):

    from dli.client.session import start_session

    return start_session(
        api_key,
        root_url=root_url,
        host=host,
        debug=debug,
        strict=strict,
        use_keyring=use_keyring
    )




"""Packaging settings."""

from os.path import abspath, dirname
from setuptools import find_packages, setup

import dli

with open('requirements.txt', 'r') as f:
    requirements = f.readlines()


setup(
    name='dli',
    python_requires='>3.6.0',
    version=dli.__version__,
    url='https://ihsmarkit.com',
    description='Data Lake command line Interface.',
    author='IHS Markit Data Lake Team',
    author_email='datalake-support@ihsmarkit.com',
    license='MOZ-2',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='cli, datalake, data, lake',
    packages=find_packages(exclude=['docs', 'tests*']),
    install_requires=requirements,
)



from datetime import timezone

from botocore.credentials import RefreshableCredentials
from botocore.session import get_session

from boto3 import Session


def create_refreshing_session(
    dli_client_session, **kwargs
):
    """
    :param dli_client_session: We must not be closing over the original
    variable in a multi-threaded env as the state can become shared.
    :param kwargs:
    :return:
    """

    def refresh():
        auth_key = dli_client_session.auth_key
        expiry_time = dli_client_session.token_expires_on.replace(
            tzinfo=timezone.utc
        ).isoformat()

        return dict(
            access_key=auth_key,
            secret_key='noop',
            token='noop',
            expiry_time=expiry_time,
        )

    _refresh_meta = refresh()

    session_credentials = RefreshableCredentials.create_from_metadata(
        metadata=_refresh_meta,
        refresh_using=refresh,
        method='noop'
    )

    session = get_session()
    session._credentials = session_credentials
    return Session(
        botocore_session=session,
        **kwargs
    )
